# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Generates the main() and call_main() functions."""

# pylint: disable=no-self-use,unused-argument,line-too-long
# flake8: noqa: E501

from ..core import get_comment
from .utils.common import main_doc_string


class ExecutionGenerator:
    """Generate the main function and its calling block for flow exporter."""

    @staticmethod
    def generate(
        content: str,
        is_async: bool,
        for_notebook: bool,
        cache_seed: int | None,
        after_run: str,
        skip_logging: bool,
    ) -> str:
        """Generate the complete flow script content.

        Parameters
        ----------
        content : str
            The content of the chats to be included in the main function.
        is_async : bool
            Whether to generate async content.
        for_notebook : bool
            Whether the export is intended for a notebook environment.
        cache_seed : str | int | None
            The cache seed to use for flow chat if any.
        after_run : str, optional
            Additional content to add after the main chat execution,
            by default ""
        skip_logging : bool, optional
            Whether to skip logging setup, by default False

        Returns
        -------
        str
            The complete flow script content.
        """
        main_function = ExecutionGenerator.generate_main_function(
            content=content,
            is_async=is_async,
            cache_seed=cache_seed,
            after_run=after_run,
            for_notebook=for_notebook,
            skip_logging=skip_logging,
        )
        call_main_function = ExecutionGenerator.generate_call_main_function(
            is_async=is_async,
            for_notebook=for_notebook,
        )
        if not for_notebook:
            execution_block = ExecutionGenerator.generate_execution_block(
                is_async=is_async,
            )
        else:
            execution_block = ""
        return (
            "\n".join([main_function, call_main_function, execution_block])
            + "\n"
        )

    @staticmethod
    def generate_store_results(is_async: bool) -> str:
        """Generate the part that writes the results to results.json.

        Parameters
        ----------
        is_async : bool
            Whether the flow is async or not.

        Returns
        -------
        str
            The part that generates the code to store the results.
        """
        content: str = "async " if is_async else ""
        tab = "    "
        content += (
            "def store_results(result_dicts: list[dict[str, Any]]) -> None:\n"
        )
        content += f'{tab}"""Store the results to results.json.\n'
        content += f"{tab}Parameters\n"
        content += f"{tab}----------\n"
        content += f"{tab}result_dicts : list[dict[str, Any]]\n"
        content += f"{tab}{tab}The list of the results.\n"
        content += f'{tab}"""\n'
        if is_async:
            content += f'{tab}async with aiofiles.open("results.json", "w", encoding="utf-8", newline="\\n") as file:\n'
            content += f"{tab}{tab}await file.write(json.dumps({{'results': result_dicts}}, indent=4, ensure_ascii=False))\n"
        else:
            content += f'{tab}with open("results.json", "w", encoding="utf-8", newline="\\n") as file:\n'
            content += f"{tab}{tab}file.write(json.dumps({{'results': result_dicts}}, indent=4, ensure_ascii=False))\n"
        return content

    @staticmethod
    def generate_store_error(is_async: bool) -> str:
        """Generate the part that writes an error to error.json.

        Parameters
        ----------
        is_async : bool
            Whether the flow is async or not.

        Returns
        -------
        str
            The content for writing the error to file.
        """
        content = "\nasync " if is_async else "\n"
        content += '''def store_error(exc: BaseException | None = None) -> None:
    """Store the error in error.json.

    Parameters
    ----------
    exc : BaseException | None
        The exception we got if any.
    """
    reason = "Event handler stopped processing" if not exc else traceback.format_exc()
    try:'''
        if is_async:
            content += """
        async with aiofiles.open("error.json", "w", encoding="utf-8", newline="\\n") as file:
            await file.write(json.dumps({"error": reason}))"""
        else:
            content += """
        with open("error.json", "w", encoding="utf-8", newline="\\n") as file:
            file.write(json.dumps({"error": reason}))"""
        content += """
    except BaseException: # pylint: disable=broad-exception-caught
        pass
"""

        return content

    @staticmethod
    def _generate_get_agent_by_name() -> str:
        """Generate the function to get an agent by its name."""
        return '''
def _get_agent_by_name(agents: list[ConversableAgent], agent_name: str) -> tuple[int, ConversableAgent | None]:
    """Get an agent by its name."""
    for ind,agent in enumerate(agents):
        if agent.name == agent_name:
            return ind, agent
    return -1, None
'''

    @staticmethod
    def _generate_handle_resume_group_pattern() -> str:
        """Generate the function to handle the detected pattern for resuming a group chat."""
        return '''
def _handle_resume_group_pattern(detected_pattern: Pattern, state_messages: list[dict[str, Any]]) -> None:
    """Handle detected pattern for resuming a group chat."""
    # pylint: disable=broad-exception-caught,too-many-try-statements
    try:
        _pattern_type = getattr(detected_pattern.__class__, "__name__", None)
    except BaseException:
        return
    if _pattern_type == "RoundRobinPattern" and state_messages:
        last_message = state_messages[-1]
        if not last_message or not isinstance(last_message, dict):
            return
        last_agent_name = last_message.get("name", "")
        if not last_agent_name:
            return
        try:
            idx, last_agent = _get_agent_by_name(detected_pattern.agents,last_agent_name)
            if last_agent and len(detected_pattern.agents)>=(idx + 1):
                detected_pattern.agents.append(detected_pattern.user_agent)
                detected_pattern.initial_agent = detected_pattern.agents[idx+1]
                detected_pattern.user_agent = detected_pattern.agents[idx]
                # fmt: off
                new_agent_order_list = detected_pattern.agents[idx+1:] + detected_pattern.agents[:idx]
                # fmt: on
                detected_pattern.agents = new_agent_order_list
        except BaseException:
            pass

'''

    @staticmethod
    def _generate_sync_prepare_resume() -> str:
        """Generate the function to prepare resuming a chat.

        Parameters
        ----------
        is_async : bool
            Whether to generate async content

        Returns
        -------
        str
            The complete function content.
        """
        content = ExecutionGenerator._generate_get_agent_by_name()
        content += ExecutionGenerator._generate_handle_resume_group_pattern()
        content += '''
def _prepare_resume(state_json: str | Path | None = None) -> None:
    """Prepare resuming a chat from state.json.

    state.json format:
        {
            "messages": [{"content": "..", "role": "...", "name": "..."}],
            "context_variables": {"key1": "value1", "key2": 4, "key3": [], "key4": {"other": "key"}}
        }
    metadata.json format:
        {
            "type": "group",
            "group": {  # one of:
                "pattern" : "<pattern_name>",
                "manager" : "<manager_name>",
            }
        }

    Parameters
    ----------
    state_json : str | Path | None
        The path to state.json to load previous state.
    """
    # pylint: disable=broad-exception-caught,too-many-try-statements,global-statement
    global __INITIAL_MSG__
    if not state_json or not Path(state_json).is_file():
        return
    metadata_json = str(state_json).replace("state.json", "metadata.json")
    if not metadata_json or not Path(metadata_json).is_file():
        return
    try:
        with open(metadata_json, "r", encoding="utf-8") as f:
            _metadata_dict = json.load(f)
    except BaseException:
        return
    if not _metadata_dict or not isinstance(_metadata_dict, dict):
        return
    _state_chat_type = _metadata_dict.get("type", "")
    if _state_chat_type != "group":
        # only resume group chats
        return
    _state_group_details = _metadata_dict.get("group", {})
    if not _state_group_details or not isinstance(_state_group_details, dict):
        return
    # either pattern or manager
    _state_group_pattern = _state_group_details.get("pattern", "")
    _state_group_manager = _state_group_details.get("manager", "")
    if not _state_group_pattern and not _state_group_manager:
        return
    try:
        with open(state_json, "r", encoding="utf-8") as f:
            _state_dict = json.load(f)
    except BaseException:
        return
    if not _state_dict or not isinstance(_state_dict, dict):
        return
    _state_messages = _state_dict.get("messages", [])
    _detected_pattern = None
    if _state_group_pattern and isinstance(_state_group_pattern, str):
        _detected_pattern = __GROUP__["patterns"].get(_state_group_pattern, None)
        if _detected_pattern:
            _state_context_variables = _state_dict.get("context_variables", {})
            if _state_context_variables and isinstance(_state_context_variables, dict):
                _detected_pattern.context_variables = ContextVariables(data=_state_context_variables)
        if _state_messages and isinstance(_state_messages, list):
            __INITIAL_MSG__ = _state_messages
    elif _state_group_manager and isinstance(_state_group_manager, str):
        _known_group_manager = __AGENTS__.get(_state_group_manager, None)
        if _known_group_manager and hasattr(_known_group_manager, "groupchat"):
            if _state_messages and isinstance(_state_messages, list):
                _known_group_manager.groupchat.messages = _state_messages
        else:
            _detected_pattern = __GROUP__["patterns"].get(f"{_state_group_manager}_pattern")
            if _detected_pattern:
                _state_context_variables = _state_dict.get("context_variables", {})
                if _state_context_variables and isinstance(_state_context_variables, dict):
                    _detected_pattern.context_variables = ContextVariables(
                        data=_state_context_variables
                    )
            if _state_messages and isinstance(_state_messages, list):
                __INITIAL_MSG__ = _state_messages
    if _detected_pattern and _state_messages and isinstance(_state_messages, list):
        _handle_resume_group_pattern(_detected_pattern, _state_messages)
'''
        return content

    @staticmethod
    def _generate_async_prepare_resume() -> str:
        """Generate the function to prepare resuming a chat (async)."""
        content = ExecutionGenerator._generate_get_agent_by_name()
        content += ExecutionGenerator._generate_handle_resume_group_pattern()
        content += '''
async def _prepare_resume(state_json: str | Path | None = None) -> None:
    """Prepare resuming a chat from state.json.

    state.json format:
        {
            "messages": [{"content": "..", "role": "...", "name": "..."}],
            "context_variables": {"key1": "value1", "key2": 4, "key3": [], "key4": {"other": "key"}}
        }
    metadata.json format:
        {
            "type": "group",
            "group": {  # one of:
                "pattern" : "<pattern_name>",
                "manager" : "<manager_name>",
            }
        }

    Parameters
    ----------
    state_json : str | Path | None
        The path to state.json to load previous state.
    """
    global __INITIAL_MSG__
    # pylint: disable=broad-exception-caught,too-many-try-statements
    if not state_json or not Path(state_json).is_file():
        return
    metadata_json = str(state_json).replace("state.json", "metadata.json")
    if not metadata_json or not Path(metadata_json).is_file():
        return
    try:
        async with aiofiles.open(metadata_json, "r", encoding="utf-8") as f:
            f_data = await f.read()
            _metadata_dict = json.loads(f_data)
    except BaseException:
        return
    if not _metadata_dict or not isinstance(_metadata_dict, dict):
        return
    _state_chat_type = _metadata_dict.get("type", "")
    if _state_chat_type != "group":
        # only resume group chats
        return
    _state_group_details = _metadata_dict.get("group", {})
    if not _state_group_details or not isinstance(_state_group_details, dict):
        return
    # either pattern or manager
    _state_group_pattern = _state_group_details.get("pattern", "")
    _state_group_manager = _state_group_details.get("manager", "")
    if not _state_group_pattern and not _state_group_manager:
        return
    try:
        async with aiofiles.open(state_json, "r", encoding="utf-8") as f:
            f_data = await f.read()
            _state_dict = json.loads(f_data)
    except BaseException:
        return
    if not _state_dict or not isinstance(_state_dict, dict):
        return
    _state_messages = _state_dict.get("context_variables", [])
    _detected_pattern = None
    if _state_group_pattern and isinstance(_state_group_pattern, str):
        _detected_pattern = __GROUP__["patterns"].get(_state_group_pattern, None)
        if _detected_pattern:
            _state_context_variables = _state_dict.get("context_variables", {})
            if _state_context_variables and isinstance(_state_context_variables, dict):
                _detected_pattern.context_variables = ContextVariables(data=_state_context_variables)
        if _state_messages and isinstance(_state_messages, list):
            __INITIAL_MSG__ = _state_messages
    elif _state_group_manager and isinstance(_state_group_manager, str):
        _known_group_manager = __AGENTS__.get(_state_group_manager, None)
        if _known_group_manager and hasattr(_known_group_manager, "groupchat"):
            if _state_messages and isinstance(_state_messages, list):
                _known_group_manager.groupchat.messages = _state_messages
        else:
            _detected_pattern = __GROUP__["patterns"].get(f"{_state_group_manager}_pattern")
            if _detected_pattern:
                _state_context_variables = _state_dict.get("context_variables", {})
                if _state_context_variables and isinstance(_state_context_variables, dict):
                    _detected_pattern.context_variables = ContextVariables(
                        data=_state_context_variables
                    )
            if _state_messages and isinstance(_state_messages, list):
                __INITIAL_MSG__ = _state_messages
    if _detected_pattern and _state_messages and isinstance(_state_messages, list):
        _handle_resume_group_pattern(_detected_pattern, _state_messages)
'''
        return content

    @staticmethod
    def generate_prepare_resume(is_async: bool) -> str:
        """Generate the function to prepare resuming a chat.

        Parameters
        ----------
        is_async : bool
            Whether to generate async content

        Returns
        -------
        str
            The complete function content.
        """
        if is_async:
            return ExecutionGenerator._generate_async_prepare_resume()
        return ExecutionGenerator._generate_sync_prepare_resume()

    @staticmethod
    def generate_main_function(
        content: str,
        is_async: bool,
        cache_seed: int | None,
        after_run: str,
        for_notebook: bool,
        skip_logging: bool,
    ) -> str:
        """Generate the main function for the flow script.

        Parameters
        ----------
        content : str
            The content of the chats to be included in the main function.
        is_async : bool
            Whether to generate async content
        cache_seed : str | int | None
            The cache seed to use for flow chat if any
        after_run : str
            Additional content to add after the main chat execution.
        for_notebook : bool
            Whether the export is intended for a notebook environment.
        skip_logging : bool, optional
            Whether to skip logging setup, by default False

        Returns
        -------
        str
            The complete main function content.
        """
        if content.startswith("\n"):
            content = content[1:]
        flow_content = "\n\n"
        comment = get_comment(
            "Start chatting",
            for_notebook=for_notebook,
        )
        flow_content += f"{comment}\n"
        if is_async:
            flow_content += "async "
        on_event_arg = "on_event: Callable[[BaseEvent, list[ConversableAgent]], bool] | None = None"
        if is_async:
            on_event_arg = (
                "on_event: "
                "Callable[[BaseEvent, list[ConversableAgent]], Coroutine[None, None, bool]]"
                " | None = None"
            )
        resume_arg = "state_json : str | Path | None = None"
        return_type_hint = "list[dict[str, Any]]"
        flow_content += (
            "def main(\n"
            f"    {on_event_arg},"
            "\n"
            f"    {resume_arg}"
            "\n"
            f") -> {return_type_hint}:"
            "\n"
        )
        flow_content += f"    {main_doc_string()}\n"
        flow_content += "    if state_json:\n"
        if is_async:
            flow_content += "        await _prepare_resume(state_json)\n"
        else:
            flow_content += "        _prepare_resume(state_json)\n"
        if not is_async:
            flow_content += "    results: list[RunResponseProtocol] | RunResponseProtocol = []\n"
        else:
            flow_content += "    results: list[AsyncRunResponseProtocol] | AsyncRunResponseProtocol = []\n"
        flow_content += "    result_dicts: list[dict[str, Any]] = []\n"
        flow_content += "    a_pause_event = asyncio.Event()\n"
        flow_content += "    a_pause_event.set()\n"
        flow_content += "    pause_event = threading.Event()\n"
        flow_content += "    pause_event.set()\n"
        space = "    "
        if cache_seed is not None:
            # noinspection SqlDialectInspection
            flow_content += (
                f"    with Cache.disk(cache_seed={cache_seed}) as cache:\n"
            )
            space = f"{space}    "
        else:
            flow_content += (
                '    if Path(".cache").is_dir():\n'
                '        shutil.rmtree(".cache", ignore_errors=True)\n'
            )
        flow_content += f"{content}" + "\n"
        if not skip_logging:
            flow_content += ExecutionGenerator._get_stop_logging_call(
                space, is_async
            )
        flow_content += "\n"
        if after_run:
            flow_content += after_run + "\n"
        if cache_seed is not None:
            space = space[4:]
        flow_content += ExecutionGenerator._get_store_results_call(
            space, is_async
        )
        flow_content += f"{space}return result_dicts\n"
        return flow_content

    @staticmethod
    def _get_stop_logging_call(space: str, is_async: bool) -> str:
        """Get stop logging call."""
        if is_async:
            return f"{space}await stop_logging()"
        return f"{space}stop_logging()"

    @staticmethod
    def _get_store_results_call(space: str, is_async: bool) -> str:
        """Get store results call."""
        if is_async:
            return f"{space}await store_results(result_dicts)\n"
        return f"{space}store_results(result_dicts)\n"

    @staticmethod
    def generate_call_main_function(is_async: bool, for_notebook: bool) -> str:
        """Generate the call_main function for the flow script.

        Parameters
        ----------
        is_async : bool
            Whether to generate async content
        for_notebook : bool
            Whether the export is intended for a notebook environment.

        Returns
        -------
        str
            The complete call_main function content.
        """
        content = "\n"
        tab = "    "
        if for_notebook:
            if is_async:
                return "# %%\nawait main()\n"
            return "# %%\nmain()\n"
        return_type_hint = "list[dict[str, Any]]"
        if is_async:
            content += "async def call_main() -> None:\n"
        else:
            content += "def call_main() -> None:\n"
        content += f'{tab}"""Run the main function and print the results."""\n'
        content += f"{tab}state_json: str | Path | None = None\n"
        content += f'{tab}if "--state" in sys.argv:\n'
        content += f'{tab}{tab}entry_index = sys.argv.index("--state")\n'
        content += f"{tab}{tab}if entry_index + 1 < len(sys.argv):\n"
        content += (
            f"{tab}{tab}{tab}state_location = Path(sys.argv[entry_index + 1])\n"
        )
        content += f"{tab}{tab}{tab}if state_location.resolve().exists():\n"
        content += f"{tab}{tab}{tab}{tab}state_json = state_location\n"
        content += f"{tab}results: {return_type_hint} = "
        if is_async:
            content += "await "
        content += "main(None, state_json=state_json)\n"
        content += f"{tab}print(json.dumps(results, default=str, indent=2, ensure_ascii=False))\n"
        return content

    @staticmethod
    def generate_execution_block(is_async: bool) -> str:
        """Generate the execution block for the main function.

        Parameters
        ----------
        is_async : bool
            Whether to generate async content

        Returns
        -------
        str
            The complete if __name__ == "__main__": block content
        """
        comment = get_comment(
            "Let's go!",
            for_notebook=False,
        )
        content = 'if __name__ == "__main__":\n'
        content += f"    {comment}"
        if is_async:
            content += "    anyio.run(call_main)\n"
        else:
            content += "    call_main()\n"
        return content


def get_result_dicts_string(space: str, is_async: bool) -> str:
    """Get the result dicts string.

    Parameters
    ----------
    space : str
        The space string to use for indentation.
    is_async : bool
        Whether the function is asynchronous.

    Returns
    -------
    str
        The result dicts string.
    """
    flow_content = f"{space}for index, result in enumerate(results):\n"
    if not is_async:
        flow_content += f"{space}    result_dict = {{\n"
        flow_content += f"{space}        'index': index,\n"
        flow_content += f"{space}        'messages': result.messages,\n"
        flow_content += f"{space}        'summary': result.summary,\n"
        flow_content += f"{space}        'cost': result.cost.model_dump(mode='json', fallback=str) if result.cost else None,\n"
        flow_content += f"{space}        'context_variables': result.context_variables.model_dump(mode='json', fallback=str) if result.context_variables else None,\n"
        flow_content += f"{space}        'last_speaker': result.last_speaker,\n"
        flow_content += f"{space}        'uuid': str(result.uuid),\n"
        flow_content += f"{space}    }}\n"
    else:
        flow_content += f"{space}    result_dict = {{\n"
        flow_content += f"{space}        'index': index,\n"
        flow_content += f"{space}        'messages': await result.messages,\n"
        flow_content += f"{space}        'summary': await result.summary,\n"
        flow_content += f"{space}        'cost': (await result.cost).model_dump(mode='json', fallback=str) if await result.cost else None,\n"
        flow_content += f"{space}        'context_variables': (await result.context_variables).model_dump(mode='json', fallback=str) if await result.context_variables else None,\n"
        flow_content += (
            f"{space}        'last_speaker': await result.last_speaker,\n"
        )
        flow_content += f"{space}        'uuid': str(result.uuid),\n"
        flow_content += f"{space}    }}\n"
    flow_content += f"{space}    result_dicts.append(result_dict)\n"
    return flow_content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught,no-self-use,too-many-try-statements
# pylint: disable=import-error,too-complex,possibly-used-before-assignment
# pyright: reportPossiblyUnboundVariable=false,reportUnknownVariableType=false
# pyright: reportUnknownMemberType=false,reportUnknownArgumentType=false
# flake8: noqa: C901

"""Filesystem-based implementation of the Storage protocol."""

import json
import os
import re
import shutil
import threading
from collections.abc import Generator, Iterable, Mapping
from contextlib import contextmanager, suppress
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from typing_extensions import Self

from .checkpoint import WaldiezCheckpoint, WaldiezCheckpointInfo
from .utils import safe_name, symlink

_PATTERNS = r"^(?!.*\.\.)(?!\.)(?!.*\.$)[\w\-.]{1,128}$"

_SAFE = re.compile(_PATTERNS, re.UNICODE)


# noinspection PyBroadException,PyUnusedLocal,TryExceptPass,PyMethodMayBeStatic
class FilesystemStorage:
    """Filesystem-based storage implementation."""

    def __init__(self, workspace_dir: Path | str = "workspace"):
        """Initialize filesystem storage.

        Parameters
        ----------
        workspace_dir : str | Path
            Base directory for all workspace data
        """
        self._workspace_dir = Path(workspace_dir).resolve()
        self._workspace_dir.mkdir(parents=True, exist_ok=True)
        self._links_registry_file = self._workspace_dir / ".links_registry.json"
        self._links_registry: dict[str, list[str]] = {}
        self._registry_lock = threading.RLock()
        self._load_links_registry()

    @staticmethod
    def load_dict(json_file: Path) -> dict[str, Any]:
        """Load dict from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.

        Returns
        -------
        dict[str, Any]
            The loaded dict.
        """
        if not json_file.is_file():
            return {}
        # pylint: disable=broad-exception-caught,too-many-try-statements
        try:
            with open(json_file, "r", encoding="utf-8") as f:
                loaded = json.load(f)
                if isinstance(loaded, dict):
                    return loaded
                if isinstance(loaded, list):
                    if len(loaded) == 1 and isinstance(loaded[0], dict):
                        return loaded[1]
                return {}
        except Exception:
            return {}

    @staticmethod
    def load_list(
        json_file: Path, fallback_dict_key: str
    ) -> list[dict[str, Any]]:
        """Load list from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.
        fallback_dict_key : str
            The key to get the list from if the loaded data is dict.

        Returns
        -------
        list[dict[str, Any]]
            The loaded list.
        """
        if not json_file.is_file():
            return []
        # pylint: disable=broad-exception-caught,too-many-try-statements
        try:
            with open(json_file, "r", encoding="utf-8") as f:
                entries = json.load(f)
            if isinstance(entries, list):
                return entries
            if isinstance(entries, dict):
                entries = entries.get(fallback_dict_key, [])
            if isinstance(entries, list):
                return entries
            return []
        except Exception:
            return []

    @property
    def workspace_dir(self) -> Path:
        """Base workspace directory."""
        return self._workspace_dir

    def save_checkpoint(
        self,
        session_name: str,
        state: dict[str, Any],
        metadata: dict[str, Any] | None = None,
        timestamp: datetime | None = None,
    ) -> Path:
        """Save a checkpoint for a session.

        Parameters
        ----------
        session_name : str
            Name of the session
        state : dict[str, Any]
            State data to save
        metadata : dict[str, Any]
            Optional metadata to include
        timestamp: datetime
            Optional timestamp (defaults to now)

        Returns
        -------
        Path
            The path of the generated checkpoint.
        """
        if timestamp is None:
            timestamp = datetime.now(timezone.utc)

        checkpoint_path = self._get_checkpoint_path(session_name, timestamp)
        checkpoint_path.mkdir(parents=True, exist_ok=True)
        state_file = checkpoint_path / "state.json"
        with open(state_file, "w", encoding="utf-8") as f:
            json.dump(state, f, indent=2, default=str)
        if metadata:
            metadata_file = checkpoint_path / "metadata.json"
            with open(metadata_file, "w", encoding="utf-8") as f:
                json.dump(metadata, f, indent=2, default=str)
        latest_link = self._get_session_dir(session_name) / "latest"
        symlink(latest_link, checkpoint_path, overwrite=True)

        return checkpoint_path

    def get_checkpoint(
        self, session_name: str, timestamp: datetime | None = None
    ) -> WaldiezCheckpointInfo | None:
        """Load a checkpoint for a session.

        Parameters
        ----------
        session_name : str
            Name of the session
        timestamp: datetime
            Optional timestamp (defaults to latest)

        Raises
        ------
        FileNotFoundError
            If no checkpoint found

        Returns
        -------
        WaldiezCheckpointInfo | None
            The loaded checkpoint info.
        """
        if timestamp is None:
            checkpoints = self._find_checkpoints(session_name)
            if not checkpoints:
                raise FileNotFoundError(
                    f"No checkpoints found for session '{session_name}'"
                )
            checkpoint = checkpoints[0]
        else:
            checkpoint_path = self._get_checkpoint_path(session_name, timestamp)
            if not checkpoint_path.exists():
                msg = (
                    f"WaldiezCheckpoint not found for session '{session_name}' "
                    f"at {WaldiezCheckpoint.format_timestamp(timestamp)}"
                )
                raise FileNotFoundError(msg)
            checkpoint = WaldiezCheckpoint(
                session_name=session_name,
                timestamp=timestamp,
                path=checkpoint_path,
            )
        return WaldiezCheckpointInfo.from_checkpoint(checkpoint)

    def load_checkpoint(
        self, info: WaldiezCheckpointInfo, history_index: int | None
    ) -> WaldiezCheckpoint:
        """Load a checkpoint.

        Parameters
        ----------
        info: WaldiezCheckpointInfo
            The checkpoint info to get the path.
        history_index: int | None
            Optional history index to use.

        Returns
        -------
        dict[str, Any]
            The loaded checkpoint data.
        """
        checkpoint = WaldiezCheckpoint(
            session_name=info.session_name,
            timestamp=info.timestamp,
            path=info.path,
        )
        if history_index is not None:
            checkpoint.load_state(history_index)
        return checkpoint

    def link_checkpoint(
        self,
        to: Path,
        session_name: str,
        timestamp: datetime | None = None,
        overwrite: bool = False,
    ) -> None:
        """Create a symlink to a checkpoint.

        Parameters
        ----------
        to: Path
            Where to create the symlink to.
        session_name : str
            The name of the session
        timestamp : datetime | None
            Optional specific timestamp
        overwrite : bool
            Overwrite existing link if needed.

        Raises
        ------
        FileNotFoundError
            If checkpoint not found
        """
        link_path = to
        if timestamp is None:
            checkpoints = self._find_checkpoints(session_name)
            if not checkpoints:
                raise FileNotFoundError(
                    f"No checkpoints found for session '{session_name}'"
                )
            checkpoint = checkpoints[0]
            link_path = to / checkpoint.path.name
        else:
            checkpoint_path = self._get_checkpoint_path(session_name, timestamp)
            if not checkpoint_path.exists():
                msg = (
                    f"WaldiezCheckpoint not found for session '{session_name}' "
                    f"at {WaldiezCheckpoint.format_timestamp(timestamp)}"
                )
                raise FileNotFoundError(msg)
            checkpoint = WaldiezCheckpoint(
                session_name=session_name,
                timestamp=timestamp,
                path=checkpoint_path,
            )
            link_path = to / checkpoint_path.name

        symlink(link_path, checkpoint.path, overwrite=overwrite)
        if not link_path.is_relative_to(self._workspace_dir):
            self._register_link(checkpoint.path, link_path)

    def list_sessions(self) -> list[str]:
        """List available sessions.

        Returns
        -------
        list[str]
            The workspace sessions.
        """
        if not self._workspace_dir.exists():
            return []
        return sorted(
            [p.name for p in self._workspace_dir.iterdir() if p.is_dir()],
            reverse=True,
        )

    def delete_session(self, session_name: str) -> None:
        """Delete a session and all its checkpoints.

        Parameters
        ----------
        session_name : str
            The session to delete.
        """
        session_dir = self._get_session_dir(session_name)
        if not session_dir.exists():
            return

        checkpoints = self._find_checkpoints(session_name)

        external_links: list[Path] = []
        for cp in checkpoints:
            external_links.extend(self._unregister_checkpoint_links(cp.path))

        for link_path in external_links:
            if link_path.is_symlink():
                try:
                    link_path.unlink(missing_ok=True)
                except Exception:
                    pass

        try:
            shutil.rmtree(session_dir)
        except Exception:
            try:
                self._clean_broken_symlinks(session_dir)
                shutil.rmtree(session_dir)
            except Exception:
                pass

        # remove any entries still pointing under this session.
        with self._registry_transaction():
            to_delete: list[str] = []
            for checkpoint_str in self._links_registry.keys():
                try:
                    rel = Path(checkpoint_str).relative_to(self._workspace_dir)
                except ValueError:
                    continue
                if rel.parts and rel.parts[0] == session_name:
                    to_delete.append(checkpoint_str)
            for key in to_delete:
                self._links_registry.pop(key, None)

    def list_checkpoints(
        self, session_name: str | None = None
    ) -> list[WaldiezCheckpointInfo]:
        """List available checkpoints.

        Parameters
        ----------
        session_name : str | None
            Optional filter by session name

        Returns
        -------
        list[WaldiezCheckpointInfo]
            The list of the checkpoints found.
        """
        checkpoints: list[WaldiezCheckpoint] = []

        if session_name:
            session_checkpoints = self._find_checkpoints(session_name)
            checkpoints.extend(session_checkpoints)
        else:
            if self._workspace_dir.exists():
                for session_dir in self._workspace_dir.iterdir():
                    if session_dir.is_dir():
                        session_checkpoints = self._find_checkpoints(
                            session_dir.name
                        )
                        if session_checkpoints:
                            checkpoints.extend(session_checkpoints)

        return [WaldiezCheckpointInfo.from_checkpoint(c) for c in checkpoints]

    def delete_checkpoint(self, session_name: str, timestamp: datetime) -> None:
        """Delete a specific checkpoint and clean up all its symlinks.

        Parameters
        ----------
        session_name : str
            The name of the session
        timestamp : datetime
            Timestamp of the checkpoint to delete

        Raises
        ------
        FileNotFoundError
            If checkpoint not found
        """
        checkpoint_path = self._get_checkpoint_path(session_name, timestamp)
        if not checkpoint_path.exists():
            msg = (
                f"WaldiezCheckpoint not found for session '{session_name}' "
                f"at {WaldiezCheckpoint.format_timestamp(timestamp)}"
            )
            raise FileNotFoundError(msg)

        external_links = self._unregister_checkpoint_links(checkpoint_path)
        for link_path in external_links:
            if link_path.is_symlink():
                try:
                    link_path.unlink(missing_ok=True)
                except Exception:
                    pass
        shutil.rmtree(checkpoint_path)
        latest_link = self._get_session_dir(session_name) / "latest"
        if latest_link.exists() and latest_link.resolve() == checkpoint_path:
            latest_link.unlink(missing_ok=True)
            # Point to next latest if available
            checkpoints = self._find_checkpoints(session_name)
            if checkpoints:
                symlink(latest_link, checkpoints[0].path, overwrite=True)

    def cleanup_old_checkpoints(
        self, session_name: str, keep_count: int = 5
    ) -> int:
        """Clean up old checkpoints, keeping only the most recent ones.

        Parameters
        ----------
        session_name : str
            Name of the session
        keep_count : int
            Number of recent checkpoints to keep

        Returns
        -------
        int
            The number of deleted checkpoints.
        """
        checkpoints = self._find_checkpoints(session_name)

        if len(checkpoints) <= keep_count:
            return 0
        checkpoints_to_delete = checkpoints[keep_count:]
        deleted_count = 0

        for checkpoint in checkpoints_to_delete:
            try:
                self.delete_checkpoint(session_name, checkpoint.timestamp)
                deleted_count += 1
            except Exception:
                pass

        return deleted_count

    # pylint: disable=too-many-branches
    def _remove_external_links(self) -> tuple[int, Iterable[str]]:
        links_to_remove: list[str] = []
        removed_count = 0
        for checkpoint_path_str, link_paths in self._links_registry.items():
            checkpoint_path = Path(checkpoint_path_str)
            if not checkpoint_path.exists():
                for link_path_str in link_paths:
                    link_path = Path(link_path_str)
                    if link_path.is_symlink():
                        try:
                            link_path.unlink(missing_ok=True)
                            removed_count += 1
                        except Exception:
                            pass
                links_to_remove.append(checkpoint_path_str)
            else:
                valid_links: list[str] = []
                for link_path_str in link_paths:
                    link_path = Path(link_path_str)
                    if link_path.is_symlink() and not link_path.exists():
                        # Broken symlink
                        try:
                            link_path.unlink(missing_ok=True)
                            removed_count += 1
                        except Exception:
                            pass
                    elif link_path.exists():
                        valid_links.append(link_path_str)

                if valid_links:
                    self._links_registry[checkpoint_path_str] = valid_links
                else:
                    links_to_remove.append(checkpoint_path_str)

        for checkpoint_path_str in links_to_remove:
            self._links_registry.pop(checkpoint_path_str, None)
        return removed_count, links_to_remove

    def clean_broken_symlinks(self, session_name: str | None = None) -> int:
        """Clean up broken symlinks.

        This cleans both internal symlinks (inside workspace) and
        registered external symlinks.

        Parameters
        ----------
        session_name : str | None
            If provided, clean only in that session's directory.
            If None, clean in all session directories.

        Returns
        -------
        int
            Total number of broken symlinks removed.
        """
        removed_count = 0
        if session_name:
            session_dir = self._get_session_dir(session_name)
            removed_count += self._clean_broken_symlinks(session_dir)
        else:
            if self._workspace_dir.exists():
                for session_dir in self._workspace_dir.iterdir():
                    if session_dir.is_dir():
                        removed_count += self._clean_broken_symlinks(
                            session_dir
                        )

        with self._registry_transaction():
            external_count, _ = self._remove_external_links()
            removed_count += external_count

        return removed_count

    def compact_registry(self) -> int:
        """Remove entries for non-existent checkpoints and links.

        Returns
        -------
        int
            Number of entries removed from registry
        """
        with self._registry_transaction():
            initial_size = len(self._links_registry)
            valid_registry = {}

            for checkpoint_str, links in self._links_registry.items():
                if Path(checkpoint_str).exists():
                    valid_links = [
                        lnk for lnk in links if Path(lnk).is_symlink()
                    ]
                    if valid_links:
                        valid_registry[checkpoint_str] = valid_links

            self._links_registry = valid_registry
            return initial_size - len(self._links_registry)

    def verify_links(
        self, session_name: str | None = None
    ) -> Mapping[str, Iterable[str]]:
        """Verify all links are valid and pointing to correct targets.

        Parameters
        ----------
        session_name : str | None
            If provided, verify only links for that session.
            If None, verify all links.

        Returns
        -------
        Mapping[str, Iterable[str]]
            Dictionary mapping checkpoint paths to list of issues found
        """
        issues: dict[str, list[str]] = {}

        with self._registry_lock:
            registry_copy = self._links_registry.copy()

        for checkpoint_str, links in registry_copy.items():
            checkpoint_path = Path(checkpoint_str)

            if session_name:
                try:
                    relative = checkpoint_path.relative_to(self._workspace_dir)
                    session_part = relative.parts[0] if relative.parts else ""
                    if session_part != session_name:
                        continue
                except ValueError:
                    continue

            for link_str in links:
                link_path = Path(link_str)

                if not link_path.exists():
                    issues.setdefault(checkpoint_str, []).append(
                        f"Missing: {link_str}"
                    )
                    continue
                if not link_path.is_symlink():
                    issues.setdefault(checkpoint_str, []).append(
                        f"Not a symlink: {link_str}"
                    )
                    continue
                try:
                    target = link_path.resolve()
                    if target != checkpoint_path:
                        issues.setdefault(checkpoint_str, []).append(
                            f"Wrong target: {link_str} -> {link_path.resolve()}"
                        )
                except Exception:
                    issues.setdefault(checkpoint_str, []).append(
                        f"Cannot resolve: {link_str}"
                    )

        return issues

    def delete_checkpoints_batch(
        self, checkpoints: Iterable[tuple[str, datetime]]
    ) -> int:
        """Delete multiple checkpoints efficiently.

        Parameters
        ----------
        checkpoints : list[tuple[str, datetime]]
            List of (session_name, timestamp) tuples

        Returns
        -------
        int
            Number of checkpoints successfully deleted
        """
        deleted = 0
        all_external_links: list[Path] = []
        checkpoints_to_delete: list[tuple[str, datetime, Path]] = []

        for session_name, timestamp in checkpoints:
            checkpoint_path = self._get_checkpoint_path(session_name, timestamp)
            if checkpoint_path.exists():
                checkpoints_to_delete.append(
                    (session_name, timestamp, checkpoint_path)
                )
                external_links = self._unregister_checkpoint_links(
                    checkpoint_path
                )
                all_external_links.extend(external_links)

        for link_path in all_external_links:
            if link_path.is_symlink():
                try:
                    link_path.unlink(missing_ok=True)
                except Exception:
                    pass
        for session_name, _, checkpoint_path in checkpoints_to_delete:
            try:
                shutil.rmtree(checkpoint_path)

                # Update latest link if needed
                latest_link = self._get_session_dir(session_name) / "latest"
                if (
                    latest_link.exists()
                    and latest_link.resolve() == checkpoint_path
                ):
                    latest_link.unlink(missing_ok=True)
                    # Point to next latest if available
                    remaining_checkpoints = self._find_checkpoints(session_name)
                    if remaining_checkpoints:
                        symlink(
                            latest_link,
                            remaining_checkpoints[0].path,
                            overwrite=True,
                        )

                deleted += 1
            except Exception:
                pass

        return deleted

    @contextmanager
    def transaction(self) -> Generator[Self, None, None]:
        """Batch multiple operations together.

        Yields
        ------
        FilesystemStorage
            Self

        Raises
        ------
        Exception
            If any of the operations fails.

        Example
        -------
        with storage.transaction():
            storage.save_checkpoint("session1", state1)
            storage.save_checkpoint("session2", state2)
            storage.delete_checkpoint("session3", timestamp)
        """
        # Backup current state
        with self._registry_lock:
            original_registry = self._links_registry.copy()

        try:
            yield self
        except Exception:
            # Rollback
            with self._registry_lock:
                self._links_registry = original_registry
                self._save_links_registry()
            raise

    @contextmanager
    def _registry_transaction(self) -> Generator[None, None, None]:
        """Context manager for safe registry operations."""
        with self._registry_lock:
            self._load_links_registry()
            yield
            self._save_links_registry()

    def _load_links_registry(self) -> None:
        """Load and validate the registry of external links."""
        if self._links_registry_file.exists():
            try:
                with open(
                    self._links_registry_file, "r", encoding="utf-8"
                ) as f:
                    data = json.load(f)
                    if isinstance(data, dict):
                        for key, value in data.items():
                            if isinstance(value, list):
                                valid_links = [
                                    v for v in value if isinstance(v, str)
                                ]
                                if valid_links:
                                    self._links_registry[key] = valid_links
            except (json.JSONDecodeError, ValueError):
                # Backup corrupted file
                backup = self._links_registry_file.with_suffix(".corrupted")
                shutil.copy2(self._links_registry_file, backup)
                self._links_registry = {}
            except Exception:
                self._links_registry = {}

    def _save_links_registry(self) -> None:
        """Save the registry with file locking."""
        temp_file = self._links_registry_file.with_suffix(".tmp")
        try:
            with open(temp_file, "w", encoding="utf-8") as f:
                json.dump(self._links_registry, f, indent=2)
            os.replace(temp_file, self._links_registry_file)
        finally:
            with suppress(FileNotFoundError):
                temp_file.unlink(missing_ok=True)

    def _register_link(self, checkpoint_path: Path, link_path: Path) -> None:
        """Register an external link in the registry."""
        key = str(checkpoint_path)
        val = str(link_path)
        with self._registry_transaction():
            self._links_registry.setdefault(key, [])
            if val not in self._links_registry[key]:
                self._links_registry[key].append(val)

    def _unregister_checkpoint_links(self, checkpoint_path: Path) -> list[Path]:
        """Get and remove all registered links for a checkpoint."""
        with self._registry_transaction():
            checkpoint_key = str(checkpoint_path)
            links = self._links_registry.pop(checkpoint_key, [])
        return [Path(link) for link in links]

    def _get_session_dir(self, session_name: str) -> Path:
        """Get the directory for a session."""
        name = safe_name(session_name)
        if not _SAFE.match(name):
            raise ValueError("Invalid session_name")
        return self._workspace_dir / name

    def _get_checkpoint_path(
        self, session_name: str, timestamp: datetime
    ) -> Path:
        """Get the path for a checkpoint."""
        timestamp_str = WaldiezCheckpoint.format_timestamp(timestamp)
        return self._get_session_dir(session_name) / timestamp_str

    def _find_checkpoints(self, session_name: str) -> list[WaldiezCheckpoint]:
        """Find all checkpoints for a session."""
        session_dir = self._get_session_dir(session_name)
        if not session_dir.exists():
            return []

        checkpoints: list[WaldiezCheckpoint] = []
        for path in session_dir.iterdir():
            if not path.is_dir():
                continue
            try:
                timestamp = WaldiezCheckpoint.parse_timestamp(path.name)
                if not timestamp:
                    continue
                state_file = path / "state.json"
                if state_file.exists():
                    checkpoint = WaldiezCheckpoint(
                        session_name=session_name,
                        timestamp=timestamp,
                        path=path,
                    )
                    checkpoints.append(checkpoint)
            except ValueError:
                continue

        return sorted(checkpoints, key=lambda c: c.timestamp, reverse=True)

    def _clean_broken_symlinks(self, directory: Path) -> int:
        """Remove broken symlinks in a directory.

        Parameters
        ----------
        directory : Path
            Directory to clean

        Returns
        -------
        int
            Number of broken symlinks removed
        """
        if not directory.exists():
            return 0

        removed = 0
        for item in directory.iterdir():
            try:
                if item.is_symlink() or item.is_dir():
                    target = item.resolve()
                    if (
                        item.is_symlink() and not item.exists()
                    ) or not target.exists():
                        item.unlink(missing_ok=True)
                        removed += 1
            except Exception:
                with suppress(Exception):
                    item.unlink(missing_ok=True)
                    removed += 1
        return removed



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
# pylint: disable=line-too-long
"""Utilities for exporting a single chat in a flow."""

from typing import Any, Callable

from waldiez.models import (
    WaldiezAgent,
    WaldiezChat,
    WaldiezChatMessage,
    WaldiezRagUserProxy,
)

from .common import get_chat_message_string, get_event_handler_string


def export_single_chat(
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    chat: WaldiezChat,
    message_kwarg: tuple[str, str],
    agent_names: dict[str, str],
    chat_names: dict[str, str],
    serializer: Callable[[str], str],
    tabs: int,
    is_async: bool,
    skip_cache: bool,
    tab_length: int = 4,
) -> tuple[str, str]:
    """Get the chat string when there is only one chat in the flow.

    Parameters
    ----------
    sender : WaldiezAgent
        The sender.
    recipient : WaldiezAgent
        The recipient.
    chat : WaldiezChat
        The chat.
    message_kwarg : tuple[str, str]
        The message kwarg and the var to use for it.
    agent_names : dict[str, str]
        A mapping of agent id to agent name.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    serializer : Callable[[str], str]
        The serializer function to escape quotes in a string.
    tabs : int
        The number of tabs to use for indentation.
    is_async : bool
        Whether the chat is asynchronous.
    skip_cache : bool
        Whether to skip the cache argument.
    tab_length : int, optional
        The length of the tab string, by default 4.

    Returns
    -------
    tuple[str, str]
        The chat string and additional methods string if any
    """
    tab = " " * tab_length * tabs if tabs > 0 else ""
    chat_args = chat.get_chat_args(for_queue=False, sender=sender)
    if not chat_args:
        return get_empty_simple_chat_string(
            chat=chat,
            sender=sender,
            recipient=recipient,
            message_kwarg=message_kwarg,
            agent_names=agent_names,
            chat_names=chat_names,
            tab=tab,
            is_async=is_async,
            skip_cache=skip_cache,
        )
    return get_simple_chat_string(
        chat=chat,
        chat_args=chat_args,
        sender=sender,
        recipient=recipient,
        message_kwarg=message_kwarg,
        agent_names=agent_names,
        chat_names=chat_names,
        serializer=serializer,
        tabs=tabs,
        is_async=is_async,
        skip_cache=skip_cache,
    )


# pylint: disable=too-many-locals
def get_simple_chat_string(
    chat: WaldiezChat,
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    message_kwarg: tuple[str, str],
    agent_names: dict[str, str],
    chat_names: dict[str, str],
    chat_args: dict[str, Any],
    serializer: Callable[..., str],
    tabs: int,
    is_async: bool,
    skip_cache: bool,
) -> tuple[str, str]:
    """Get the chat string when there are chat arguments.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    sender : WaldiezAgent
        The sender.
    recipient : WaldiezAgent
        The recipient.
    message_kwarg : tuple[str, str]
        The message kwarg and the var to use for it.
    agent_names : dict[str, str]
        A mapping of agent id to agent name.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    chat_args : dict[str, Any]
        The chat arguments.
    serializer : Callable[[str], str]
        The serializer function to escape quotes in a string.
    tabs : int
        The number of tabs to use for indentation.
    is_async : bool
        Whether the chat is asynchronous.
    skip_cache : bool
        Whether to skip the cache argument.

    Returns
    -------
    tuple[str, str]
        The chat string and additional methods string if any.
    """
    message_arg, message_var = message_kwarg
    tab = "    " * tabs
    sender_name = agent_names[sender.id]
    initiate = "run"
    if is_async:
        sender_name = f"await {sender_name}"
        initiate = "a_run"
    recipient_name = agent_names[recipient.id]
    chat_string = "\n" + f"{tab}results = {sender_name}.{initiate}(" + "\n"
    chat_string += f"{tab}    {recipient_name},"
    if not skip_cache:
        chat_string += "\n" + f"{tab}    cache=cache,"
    for key, value in chat_args.items():
        if key == message_arg:
            chat_string += "\n" + f"{tab}    {key}={message_var},"
            continue
        if isinstance(value, str):
            chat_string += "\n" + f'{tab}    {key}="{value}",'
        elif isinstance(value, dict):
            chat_string += (
                "\n" + f"{tab}    {key}={serializer(value, tabs=tabs + 1)},"
            )
        else:
            chat_string += "\n" + f"{tab}    {key}={value},"
    message_arg, before_chat = _get_chat_message(
        tab=tab,
        chat=chat,
        chat_names=chat_names,
        sender=sender,
        sender_name=sender_name,
        message_kwarg=message_kwarg,
    )
    chat_string += message_arg
    summary_arg, before_summary = _get_chat_summary(
        tab=tab,
        chat=chat,
        chat_names=chat_names,
    )
    chat_string += summary_arg
    chat_string += "\n" + f"{tab})" + "\n"
    chat_string += get_event_handler_string(space=tab, is_async=is_async)
    if before_summary:
        before_chat += "\n" + before_summary
    return chat_string, before_chat


def get_empty_simple_chat_string(
    chat: WaldiezChat,
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    message_kwarg: tuple[str, str],
    agent_names: dict[str, str],
    chat_names: dict[str, str],
    tab: str,
    is_async: bool,
    skip_cache: bool,
) -> tuple[str, str]:
    """Get the chat string when there are no chat arguments.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    sender : WaldiezAgent
        The sender.
    recipient : WaldiezAgent
        The recipient.
    message_kwarg : tuple[str, str]
        The message kwarg and the var to use for it.
    agent_names : dict[str, str]
        A mapping of agent id to agent name.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    tab : str
        The tab string.
    is_async : bool
        Whether the chat is asynchronous.
    skip_cache : bool
        Whether to skip the cache argument.

    Returns
    -------
    tuple[str, str]
        The chat string and additional methods string if any
    """
    sender_name = agent_names[sender.id]
    if is_async:
        sender_name = f"await {sender_name}"
    recipient_name = agent_names[recipient.id]
    initiate = "a_run" if is_async else "run"
    content = "\n" + f"{tab}results = {sender_name}.{initiate}(" + "\n"
    content += f"{tab}    {recipient_name}," + "\n"
    if not skip_cache:
        content += f"{tab}    cache=cache," + "\n"
    message_arg, before_message = _get_chat_message(
        tab=tab,
        chat=chat,
        chat_names=chat_names,
        sender=sender,
        sender_name=sender_name,
        message_kwarg=message_kwarg,
    )
    content += message_arg
    summary_arg, before_summary = _get_chat_summary(
        tab=tab,
        chat=chat,
        chat_names=chat_names,
    )
    content += summary_arg
    content += f"{tab})" + "\n"
    content += get_event_handler_string(space=tab, is_async=is_async)
    before_chat = before_message + before_summary
    return content, before_chat


def _get_chat_message(
    tab: str,
    chat: WaldiezChat,
    chat_names: dict[str, str],
    sender: WaldiezAgent,
    sender_name: str,
    message_kwarg: tuple[str, str],
) -> tuple[str, str]:
    """Get the chat message string.

    Parameters
    ----------
    tab : str
        The tab string.
    chat : WaldiezChat
        The chat.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    sender : WaldiezAgent
        The sender.
    sender_name : str
        The sender name.
    message_kwarg : tuple[str, str]
        The message kwarg and the var to use for it.

    Returns
    -------
    tuple[str, str]
        The message argument and additional methods string if any.
    """
    before_chat = ""
    message_arg, message_var = message_kwarg
    message_arg_var = "\n" + f"{tab}    {message_arg}={message_var},"
    if (
        sender.is_rag_user
        and isinstance(sender, WaldiezRagUserProxy)
        and chat.message.type == "rag_message_generator"
        and chat.message.use_carryover is False
    ):
        before_chat = f"{message_var}={sender_name}.message_generator"
        return message_arg_var, before_chat
    message_value, method_content = get_chat_message_string(
        sender=sender,
        chat=chat,
        chat_names=chat_names,
    )
    if message_value and isinstance(chat.data.message, WaldiezChatMessage):
        if chat.data.message.type in ("method", "rag_message_generator"):
            before_chat = method_content if method_content else ""
            before_chat += "\n"
            before_chat += f"{message_var}={message_value}"
            return message_arg_var, before_chat
        if chat.message.type == "string" and chat.data.message.content:
            before_chat += f"{message_var}={message_value}"
            return message_arg_var, before_chat
        if chat.message.type == "none":
            before_chat += f"{message_var}=None"
            return message_arg_var, before_chat
    return "", before_chat  # pragma: no cover


def _get_chat_summary(
    tab: str,
    chat: WaldiezChat,
    chat_names: dict[str, str],
) -> tuple[str, str]:
    """Get the chat summary string.

    Parameters
    ----------
    tab : str
        The tab string.
    chat : WaldiezChat
        The chat.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.

    Returns
    -------
    tuple[str, str]
        The message argument and additional methods string if any.
    """
    before_summary = ""
    summary_arg = ""
    if chat.summary.method == "custom":
        chat_name = chat_names[chat.id]
        summary_source, summary_value = chat.get_summary_function(
            name_suffix=chat_name,
        )
        if summary_value and summary_source:
            summary_arg = "\n" + f"{tab}    summary_method={summary_value},"
            before_summary = summary_source

    return summary_arg, before_summary  # pragma: no cover



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Predefined tools registry for Waldiez."""

from ._config import PredefinedToolConfig
from ._duckduckgo import DuckDuckGoSearchConfig
from ._google import GoogleSearchConfig
from ._perplexity import PerplexitySearchConfig
from ._searxng import SearxNGSearchConfig
from ._tavily import TavilySearchConfig
from ._wikipedia import WikipediaSearchConfig
from ._youtube import YouTubeSearchConfig

PREDEFINED_TOOLS: dict[str, PredefinedToolConfig] = {
    GoogleSearchConfig.name: GoogleSearchConfig,
    WikipediaSearchConfig.name: WikipediaSearchConfig,
    YouTubeSearchConfig.name: YouTubeSearchConfig,
    TavilySearchConfig.name: TavilySearchConfig,
    DuckDuckGoSearchConfig.name: DuckDuckGoSearchConfig,
    PerplexitySearchConfig.name: PerplexitySearchConfig,
    SearxNGSearchConfig.name: SearxNGSearchConfig,
}


def get_predefined_tool_config(tool_name: str) -> PredefinedToolConfig | None:
    """Get configuration for a predefined tool.

    Parameters
    ----------
    tool_name : str
        Name of the tool to retrieve configuration for.

    Returns
    -------
    PredefinedToolConfig | None
        Configuration of the tool if it exists, otherwise None.
    """
    return PREDEFINED_TOOLS.get(tool_name)


def get_predefined_tool_requirements(tool_name: str) -> list[str]:
    """Get requirements for a predefined tool.

    Parameters
    ----------
    tool_name : str
        Name of the tool to retrieve requirements for.

    Returns
    -------
    list[str]
        List of requirements for the tool,
        or an empty list if the tool does not exist.
    """
    config = get_predefined_tool_config(tool_name)
    return config.requirements if config else []


def get_predefined_tool_imports(tool_name: str) -> list[str]:
    """Get imports for a predefined tool.

    Parameters
    ----------
    tool_name : str
        Name of the tool to retrieve imports for.

    Returns
    -------
    list[str]
        List of imports for the tool,
        or an empty list if the tool does not exist.
    """
    config = get_predefined_tool_config(tool_name)
    return config.implementation.tool_imports if config else []


def list_predefined_tools() -> list[str]:
    """List all available predefined tools.

    Returns
    -------
    list[str]
        List of all predefined tool names.
    """
    return list(PREDEFINED_TOOLS.keys())



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
# pyright: reportReturnType=false
"""Core exporting protocols."""

from typing import Any, Protocol, runtime_checkable

from .result import ExportResult
from .validation import ValidationResult


# Protocol for extras that can contribute to exports
@runtime_checkable
class ExportContributor(Protocol):
    """Protocol for objects that can contribute to exports."""

    def contribute_to_export(self, result: ExportResult) -> None:
        """Contribute content to the export result.

        Parameters
        ----------
        result : ExportResult
            The export result to contribute to.
        """


# Protocol for validation
class Validator(Protocol):
    """Protocol for validation components."""

    def validate(self, content: Any) -> ValidationResult:
        """Validate content and return result.

        Parameters
        ----------
        content : Any
            The content to validate.

        Returns
        -------
        ValidationResult
            The validation result.
        """


# Protocol for serialization
class Serializer(Protocol):
    """Protocol for serialization components."""

    def serialize(self, obj: Any, **kwargs: Any) -> str:
        """Serialize an object to string representation.

        Parameters
        ----------
        obj : Any
            The object to serialize.
        **kwargs : Any
            Additional keyword arguments for serialization.

        Returns
        -------
        str
            The serialized representation.
        """


# Protocol for resolving a path
class PathResolver(Protocol):
    """Protocol for resolving a path."""

    def resolve(self, path: str) -> str:
        """Resolve a path.

        Parameters
        ----------
        path : str
            The path to resolve.

        Returns
        -------
        str
            The resolved path.

        Raises
        ------
        ValueError
            If the path cannot be resolved.
        """

    def is_local(self, path: str) -> bool:
        """Check if the given path is a local path.

        Parameters
        ----------
        path : str
            The path to check.

        Returns
        -------
        bool
            True if the path is a local path, False otherwise.
        """


class ContentGenerator(Protocol):
    """Protocol for generating content."""

    def generate(
        self,
        merged_result: ExportResult,
        is_async: bool,
        after_run: str,
        skip_logging: bool,
        **kwargs: Any,
    ) -> str:
        """Generate content based on provided parameters.

        Parameters
        ----------
        merged_result : ExportResult
            The merged export result containing all content.
        is_async : bool
            Whether to generate async content.
        after_run : str
            Additional content to add after the main flow execution.
        skip_logging : bool
            Whether to skip logging setup.
        **kwargs : Any
            Parameters to influence content generation.

        Returns
        -------
        str
            The generated content.


        Raises
        ------
        ExporterContentError
            If exporting fails.
        """



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent exporter."""

from .exporter import AgentExporter
from .factory import create_agent_exporter

__all__ = [
    "AgentExporter",
    "create_agent_exporter",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Custom exceptions for Waldiez runners."""


class StopRunningException(Exception):
    """Exception to stop the running process."""

    reason: str = "Execution stopped by user"

    def __str__(self) -> str:
        """Get the string representation of the exception.

        Returns
        -------
        str
            The string representation of the exception.
        """
        return self.reason

    def __repr__(self) -> str:
        """Get the string representation of the exception.

        Returns
        -------
        str
            The string representation of the exception.
        """
        return self.reason



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportMissingImports=false,reportUnknownVariableType=false
# pyright: reportPossiblyUnboundVariable=false,reportUnknownMemberType=false
# pyright: reportUnknownParameterType=false,reportUnknownArgumentType=false
# pyright: reportAttributeAccessIssue=false,reportAny=false
# pyright: reportConstantRedefinition=false,reportAssignmentType=false
# pyright: reportImportCycles=false,reportGeneralTypeIssues=false
# pylint: disable=import-error,line-too-long,invalid-name
# flake8: noqa: E501
"""WebSocket server implementation for Waldiez."""

import asyncio
import logging
import re
import signal
import time
import traceback
import uuid
from collections.abc import Sequence
from pathlib import Path
from typing import Any, final

from .client_manager import ClientManager
from .errors import ErrorHandler, MessageParsingError, ServerOverloadError
from .models import ConnectionNotification
from .session_manager import SessionManager
from .utils import get_available_port, is_port_available

HAS_WATCHDOG = False
try:
    from .reloader import FileWatcher, create_file_watcher

    HAS_WATCHDOG = True
except ImportError:

    class FileWatcher:  # type: ignore[no-redef]
        """Mock fileWatcher for typing."""

        def start(self) -> None:
            """Start the watcher."""

        def stop(self) -> None:
            """Stop the watcher."""

    # pylint: disable=missing-param-doc,missing-return-doc
    def create_file_watcher(*args: Any, **kwargs: Any) -> FileWatcher:  # type: ignore
        """No file watcher available."""
        return FileWatcher(*args, **kwargs)


HAS_WEBSOCKETS = False
try:
    import websockets  # type: ignore[unused-ignore, unused-import, import-not-found, import-untyped]
    from websockets.exceptions import (  # type: ignore[unused-ignore, unused-import, import-not-found, import-untyped]
        ConnectionClosed,
        WebSocketException,
    )

    HAS_WEBSOCKETS = True
except ImportError:
    from ._mock import (  # type: ignore[no-redef, unused-ignore, unused-import, import-not-found, import-untyped]
        websockets,
    )

    ConnectionClosed = websockets.ConnectionClosed  # type: ignore[no-redef,unused-ignore,unused-import,import-not-found,import-untyped,misc]
    WebSocketException = websockets.WebSocketException  # type: ignore[no-redef,unused-ignore,unused-import,import-not-found,import-untyped,misc]


logger = logging.getLogger(__name__)

CWD = Path.cwd()


# pylint: disable=too-many-instance-attributes
@final
class WaldiezWsServer:
    """WebSocket server for Waldiez."""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 8765,
        auto_reload: bool = False,
        workspace_dir: Path = CWD,
        max_clients: int = 2,
        allowed_origins: Sequence[re.Pattern[str]] | None = None,
        **kwargs: Any,
    ):
        """Initialize WebSocket server.

        Parameters
        ----------
        host : str
            Server host address
        port : int
            Server port
        auto_reload : bool
            Enable automatic reloading of the server on code changes
        workspace_dir : Path
            Path to the workspace directory
        max_clients : int
            Maximum number of concurrent clients (default: 2)
        allowed_origins : Sequence[re.Pattern[str]] | None
            List of allowed origins for CORS (default: None)
        ping_interval : float | None
            Ping interval in seconds
        ping_timeout : float | None
            Ping timeout in seconds
        close_timeout : float | None
            Close timeout in seconds
        max_size : int | None
            Maximum message size in bytes
        max_queue : int | None
            Maximum queue size
        write_limit : int
            Write buffer limit
        """
        self.host = host
        self.port = port
        self.auto_reload = auto_reload and HAS_WATCHDOG
        self.workspace_dir = workspace_dir
        self.max_clients = max_clients
        self.allowed_origins = allowed_origins

        # WebSocket configuration
        self.ping_interval = kwargs.get("ping_interval", 20.0)
        self.ping_timeout = kwargs.get("ping_timeout", 20.0)
        self.close_timeout = kwargs.get("close_timeout", 10.0)
        self.max_size = kwargs.get("max_size", 2**23)  # 8MB
        self.max_queue = kwargs.get("max_queue", 32)
        self.write_limit = kwargs.get("write_limit", 2**16)  # 64KB

        # Server state
        self.server: websockets.Server | None = None
        self.session_manager = SessionManager()
        self.clients: dict[str, ClientManager] = {}
        self.is_running = False
        self.start_time = 0.0
        self.error_handler = ErrorHandler()

        # Shutdown event
        self.shutdown_event = asyncio.Event()

        # Statistics
        self.stats = {
            "connections_total": 0,
            "connections_active": 0,
            "messages_received": 0,
            "messages_sent": 0,
        }

    # pylint: disable=too-complex,too-many-branches,too-many-statements
    async def _handle_client(  # noqa: C901
        self,
        websocket: websockets.ServerConnection,
    ) -> None:
        """Handle individual client connections.

        Parameters
        ----------
        websocket : websockets.WebSocketServerProtocol
            WebSocket connection
        """
        client_id = str(uuid.uuid4())

        # Check client limit
        if len(self.clients) >= self.max_clients:
            logger.warning(
                "Client limit exceeded (%d/%d), rejecting connection from %s",
                len(self.clients),
                self.max_clients,
                websocket.remote_address,
            )
            error = ServerOverloadError(len(self.clients), self.max_clients)
            await websocket.close(code=error.error_code, reason=error.message)
            return

        # Create client handler
        client_manager = ClientManager(
            websocket,
            client_id,
            self.session_manager,
            workspace_dir=self.workspace_dir,
            error_handler=self.error_handler,
        )
        self.clients[client_id] = client_manager
        self.stats["connections_total"] += 1
        self.stats["connections_active"] = len(self.clients)
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            await client_manager.send_message(
                ConnectionNotification(
                    status="connected",
                    client_id=client_id,
                    server_time=time.time(),
                ).model_dump(mode="json", fallback=str)
            )

            # Message handling loop
            # noinspection PyTypeChecker
            async for raw_message in websocket:
                try:
                    # Parse message
                    if isinstance(raw_message, bytes):
                        message_str = raw_message.decode(
                            "utf-8", errors="replace"
                        )
                    else:
                        # noinspection PyUnreachableCode
                        message_str = (
                            raw_message
                            if isinstance(raw_message, str)
                            else str(raw_message)
                        )
                    response = await client_manager.handle_message(message_str)
                    self.stats["messages_received"] += 1

                    # Send response if available
                    if response:  # pragma: no branch
                        success = await client_manager.send_message(response)
                        if success:
                            self.stats["messages_sent"] += 1
                        else:
                            self.error_handler.record_send_failure(client_id)

                except ValueError as e:
                    logger.warning("Invalid message from %s: %s", client_id, e)
                    error_response = self.error_handler.handle_error(
                        MessageParsingError(str(e), str(raw_message)),
                        client_id=client_id,
                    )
                    await client_manager.send_message(error_response)

                except Exception as e:
                    traceback.print_exc()
                    logger.error(
                        "Error handling message from %s: %s", client_id, e
                    )
                    error_response = self.error_handler.handle_error(
                        e, client_id=client_id
                    )
                    await client_manager.send_message(error_response)

        except ConnectionClosed:
            logger.info("Client %s disconnected normally", client_id)

        except WebSocketException as e:
            logger.warning("WebSocket error for client %s: %s", client_id, e)
            self.error_handler.record_operational_error(
                "WebSocketException", str(e)
            )

        except Exception as e:
            logger.error(
                "Unexpected error handling client %s: %s", client_id, e
            )
            self.error_handler.record_operational_error(
                "UnexpectedError", str(e)
            )
            raise

        finally:
            # Clean up client
            if client_id in self.clients:  # pragma: no branch
                self.clients[client_id].close_connection()
                del self.clients[client_id]
                self.stats["connections_active"] = len(self.clients)

    async def start(self) -> None:
        """Start the WebSocket server.

        Raises
        ------
        RuntimeError
            If the port is already in use
        Exception
            For any other errors
        """
        if self.is_running:
            logger.warning("Server is already running")
            return

        await self.session_manager.start()
        # Check port availability
        if not self.auto_reload and not is_port_available(self.port):
            logger.warning("Port %d is not available", self.port)
            self.port = get_available_port()
            logger.info("Using port %d", self.port)

        logger.info("Starting WebSocket server on %s:%d", self.host, self.port)
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            # Create server
            # noinspection PyTypeChecker
            self.server = await websockets.serve(
                self._handle_client,
                self.host,
                self.port,
                ping_interval=self.ping_interval,
                ping_timeout=self.ping_timeout,
                close_timeout=self.close_timeout,
                max_size=self.max_size,
                max_queue=self.max_queue,
                write_limit=self.write_limit,
                origins=self.allowed_origins,
                # Additional settings
                compression=None,  # Disable compression for lower latency
                logger=logger,
                server_header="Waldiez/ws",
            )

            self.is_running = True
            self.start_time = time.time()

            logger.info("WebSocket server started successfully")
            logger.info("Server configuration:")
            logger.info("  - Host: %s", self.host)
            logger.info("  - Port: %d", self.port)
            logger.info("  - Max clients: %d", self.max_clients)
            logger.info("  - Ping interval: %s", self.ping_interval)
            logger.info("  - Max message size: %s", self.max_size)

            # Wait for shutdown
            await self.shutdown_event.wait()

        except Exception as e:
            logger.error("Failed to start server: %s", e)
            raise

        finally:
            await self.stop()

    async def stop(self) -> None:
        """Stop the WebSocket server."""
        await self.session_manager.stop()
        if not self.is_running:
            logger.warning("Server is not running")
            return

        logger.info("Stopping WebSocket server...")

        # Close all client connections
        if self.clients:
            logger.info(
                "Closing %d active client connections", len(self.clients)
            )
            close_tasks: list[Any] = []
            for client in self.clients.values():
                if client.is_active:  # pragma: no branch
                    close_tasks.append(client.websocket.close())

            if close_tasks:  # pragma: no branch
                await asyncio.gather(*close_tasks, return_exceptions=True)

        # Stop server
        if self.server:
            self.server.close()
            await self.server.wait_closed()

        self.is_running = False
        self.clients.clear()
        self.stats["connections_active"] = 0

        uptime = time.time() - self.start_time
        logger.info("WebSocket server stopped (uptime: %.1f seconds)", uptime)

    def shutdown(self) -> None:
        """Trigger server shutdown."""
        self.shutdown_event.set()

    def get_stats(self) -> dict[str, Any]:
        """Get server statistics.

        Returns
        -------
        dict[str, Any]
            Server statistics
        """
        uptime = time.time() - self.start_time if self.is_running else 0
        return {
            **self.stats,
            "uptime_seconds": uptime,
            "is_running": self.is_running,
            "server_config": {
                "host": self.host,
                "port": self.port,
                "max_clients": self.max_clients,
                "ping_interval": self.ping_interval,
                "max_size": self.max_size,
            },
            "error_stats": self.error_handler.get_error_stats(),
        }

    async def broadcast(
        self, message: dict[str, Any], exclude_client: str | None = None
    ) -> int:
        """Broadcast message to all connected clients.

        Parameters
        ----------
        message : Dict[str, Any]
            Message to broadcast
        exclude_client : Optional[str]
            Client ID to exclude from broadcast

        Returns
        -------
        int
            Number of clients that received the message
        """
        if not self.clients:
            return 0

        send_tasks: list[Any] = []
        for client_id, client in self.clients.items():
            if client_id != exclude_client and client.is_active:
                send_tasks.append(client.send_message(message))

        if not send_tasks:  # pragma: no cover
            return 0

        results = await asyncio.gather(*send_tasks, return_exceptions=True)
        successful = sum(1 for result in results if result is True)

        self.stats["messages_sent"] += successful
        return successful


def _get_file_watcher(
    auto_reload: bool, watch_dirs: set[Path] | None
) -> FileWatcher | None:
    file_watcher: FileWatcher | None = None
    if auto_reload and HAS_WATCHDOG:
        # pylint: disable=import-outside-toplevel,too-many-try-statements
        try:
            # Determine watch directories
            if watch_dirs is None:
                project_root = Path(__file__).parents[2]

                # Watch the actual waldiez package directory
                waldiez_dir = project_root / "waldiez"
                if waldiez_dir.exists():
                    watch_dirs = {waldiez_dir}
                    logger.info(
                        "Auto-reload: watching waldiez package at %s",
                        waldiez_dir,
                    )
                else:
                    # Fallback: watch current directory
                    watch_dirs = {Path.cwd()}
                    logger.warning(
                        "Auto-reload: fallback to current directory %s",
                        Path.cwd(),
                    )

            # Create file watcher with restart callback
            file_watcher = create_file_watcher(
                root_dir=Path(__file__).parents[2],
                additional_dirs=list(watch_dirs),
                restart_callback=None,
            )
            file_watcher.start()
            logger.info(
                "Auto-reload enabled for directories: %s",
                {str(dir_) for dir_ in watch_dirs},
            )

        except ImportError as e:
            logger.warning("Auto-reload not available: %s", e)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("Failed to set up auto-reload: %s", e)
    return file_watcher


async def run_server(
    host: str = "localhost",
    port: int = 8765,
    workspace_dir: Path = CWD,
    auto_reload: bool = False,
    watch_dirs: set[Path] | None = None,
    **server_kwargs: Any,
) -> None:
    """Run the WebSocket server with optional auto-reload.

    Parameters
    ----------
    host : str
        Server host
    port : int
        Server port
    workspace_dir : Path
        Path to the workspace directory
    auto_reload : bool
        Enable auto-reload on file changes
    watch_dirs : Optional[Set[Path]]
        Directories to watch for auto-reload
    **server_kwargs
        Additional server configuration
    """
    server = WaldiezWsServer(
        host=host,
        port=port,
        auto_reload=auto_reload,
        workspace_dir=workspace_dir,
        **server_kwargs,
    )

    # Set up signal handlers
    def signal_handler() -> None:
        """Handle shutdown signals."""
        logger.info("Received shutdown signal")
        server.shutdown()
        logger.info("Shutdown event set, stopping server...")

    # Register signal handlers
    loop = asyncio.get_running_loop()
    try:
        for sig in (signal.SIGTERM, signal.SIGINT):
            # noinspection PyTypeChecker
            loop.add_signal_handler(sig, signal_handler)
    except NotImplementedError:
        # Fallback for Windows
        signal.signal(signal.SIGINT, lambda s, f: signal_handler())
        signal.signal(signal.SIGTERM, lambda s, f: signal_handler())

    # Set up auto-reload if requested
    file_watcher = _get_file_watcher(
        auto_reload=auto_reload, watch_dirs=watch_dirs
    )
    try:
        # Start server
        await server.start()
    finally:
        # Clean up file watcher
        if file_watcher:
            file_watcher.stop()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportAttributeAccessIssue=false,reportUnknownArgumentType=false
# flake8: noqa: G004
"""Waldiez subprocess runner that inherits from BaseRunner."""

import asyncio
import re
from pathlib import Path
from typing import Any, Callable, Literal

from typing_extensions import override

from waldiez.models import Waldiez

from ..base_runner import WaldiezBaseRunner
from ..step_by_step.breakpoints_mixin import BreakpointsMixin
from ._async_runner import AsyncSubprocessRunner
from ._sync_runner import SyncSubprocessRunner


# noinspection PyUnusedLocal,PyBroadException
class WaldiezSubprocessRunner(WaldiezBaseRunner):
    """Waldiez runner that uses subprocess execution via standalone runners."""

    def __init__(
        self,
        waldiez: Waldiez,
        on_output: Callable[[dict[str, Any]], None] | None = None,
        on_input_request: Callable[[str], None] | None = None,
        on_async_output: Callable[[dict[str, Any]], Any] | None = None,
        on_async_input_request: Callable[[str], Any] | None = None,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool = True,
        dot_env: str | Path | None = None,
        input_timeout: float = 120.0,
        **kwargs: Any,
    ) -> None:
        """Initialize subprocess runner that inherits from BaseRunner.

        Parameters
        ----------
        waldiez : Waldiez
            The Waldiez workflow to run
        on_output : Callable[[dict[str, Any]], None] | None
            Sync callback for handling output messages
        on_input_request : Callable[[str], None] | None
            Sync callback for handling input requests
        on_async_output : Callable[[dict[str, Any]], Any] | None
            Async callback for handling output messages
        on_async_input_request : Callable[[str], Any] | None
            Async callback for handling input requests
        output_path : str | Path | None
            Output path for the workflow
        uploads_root : str | Path | None
            Root directory for uploads
        structured_io : bool
            Whether to use structured I/O (forced to True for subprocess)
        dot_env : str | Path | None
            Path to .env file
        input_timeout : float
            Timeout for user input in seconds
        **kwargs : Any
            Additional arguments for BaseRunner
        """
        super().__init__(
            waldiez=waldiez,
            output_path=output_path,
            uploads_root=uploads_root,
            structured_io=True,  # Always force structured I/O for subprocess
            dot_env=dot_env,
            **kwargs,
        )
        self.breakpoints = self._parse_breakpoints(**kwargs)
        # Store callbacks
        self.sync_on_output = on_output or self._default_sync_output
        self.sync_on_input_request = (
            on_input_request or self._default_sync_input_request
        )
        self.async_on_output = on_async_output or self._default_async_output
        self.async_on_input_request = (
            on_async_input_request or self._default_async_input_request
        )
        self.input_timeout = input_timeout

        # Subprocess runner instances
        self.async_runner: AsyncSubprocessRunner | None = None
        self.sync_runner: SyncSubprocessRunner | None = None
        self.temp_flow_file: Path | None = None
        mode = kwargs.get("mode", "run")
        if mode not in ["run", "debug"]:
            raise ValueError(f"Invalid mode: {mode}")
        # noinspection PyTypeChecker
        self.mode: Literal["run", "debug"] = mode
        waldiez_file = kwargs.get("waldiez_file")
        self._waldiez_file = self._ensure_waldiez_file(waldiez_file)

    def _ensure_waldiez_file(self, waldiez_file: str | Path | None) -> Path:
        """Ensure the Waldiez file is a Path object."""
        if isinstance(waldiez_file, str):
            waldiez_file = Path(waldiez_file)
        if waldiez_file and waldiez_file.is_file():
            return waldiez_file.resolve()
        file_name = self.waldiez.name
        # sanitize file name
        # noinspection RegExpRedundantEscape
        file_name = re.sub(r"[^a-zA-Z0-9_\-\.]", "_", file_name)[:30]
        file_name = f"{file_name}.waldiez"
        with open(file_name, "w", encoding="utf-8", newline="\n") as f:
            f.write(self.waldiez.model_dump_json())
        return Path(file_name).resolve()

    @staticmethod
    def _parse_breakpoints(**kwargs: Any) -> list[str]:
        initial_breakpoints = kwargs.get("breakpoints")
        if isinstance(initial_breakpoints, (list, set, tuple)):
            breakpoints = BreakpointsMixin.get_initial_breakpoints(
                initial_breakpoints
            )
            return [str(item) for item in breakpoints]
        return []

    def _default_sync_output(self, data: dict[str, Any]) -> None:
        """Get the default sync output handler."""
        if data.get("type") == "error":
            self.log.error(data.get("data", ""))
        else:
            content = data.get("data", data)
            self.print(content)

    def _default_sync_input_request(self, prompt: str) -> None:
        """Get the default sync input request handler."""
        input_value = input(prompt)
        self.log.debug("User input received: %s", input_value)
        self.provide_user_input(input_value)

    async def _default_async_output(self, data: dict[str, Any]) -> None:
        """Get the default async output handler."""
        self._default_sync_output(data)

    async def _default_async_input_request(self, prompt: str) -> None:
        """Get the default async input request handler."""
        await asyncio.to_thread(self._default_sync_input_request, prompt)

    def _create_async_subprocess_runner(self) -> AsyncSubprocessRunner:
        """Create async subprocess runner."""
        self.async_runner = AsyncSubprocessRunner(
            on_output=self.async_on_output,
            on_input_request=self.async_on_input_request,
            input_timeout=self.input_timeout,
            uploads_root=self.uploads_root,
            dot_env=self.dot_env_path,
            logger=self.log,
            breakpoints=self.breakpoints,
            checkpoint=WaldiezBaseRunner._checkpoint,
        )
        return self.async_runner

    def _create_sync_subprocess_runner(self) -> SyncSubprocessRunner:
        """Create sync subprocess runner."""
        self.sync_runner = SyncSubprocessRunner(
            on_output=self.sync_on_output,
            on_input_request=self.sync_on_input_request,
            input_timeout=self.input_timeout,
            uploads_root=self.uploads_root,
            dot_env=self.dot_env_path,
            logger=self.log,
            breakpoints=self.breakpoints,
            checkpoint=WaldiezBaseRunner._checkpoint,
        )
        return self.sync_runner

    @override
    @staticmethod
    def gather() -> tuple[bool, str]:
        """Gather any results after run.

        Returns
        -------
        tuple[bool, str]
            True if operation succeeded and any related message.
        """
        return SyncSubprocessRunner.gather()

    @staticmethod
    async def a_gather() -> tuple[bool, str]:
        """Gather any results after run.

        Returns
        -------
        tuple[bool, str]
            True if operation succeeded and any related message.
        """
        return await AsyncSubprocessRunner.gather()

    @override
    def run(
        self,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow.

        Parameters
        ----------
        output_path : str | Path | None
            The output path, by default None.
        uploads_root : str | Path | None
            The runtime uploads root, by default None.
        structured_io : bool
            Whether to use structured IO instead of the default 'input/print'.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool
            Whether to skip generating the timeline JSON.
        dot_env : str | Path | None
            The path to the .env file, if any.
        **kwargs : Any
            Additional keyword arguments for the run method.

        Returns
        -------
        list[dict[str, Any]]
            The results of the run.
        """
        temp_dir = Path.cwd()
        output_file = self._get_output_file(output_path)
        if dot_env is not None:  # pragma: no cover
            resolved = Path(dot_env).resolve()
            if resolved.is_file():
                WaldiezBaseRunner._dot_env_path = resolved
        return self._run(
            temp_dir=temp_dir,
            output_file=output_file,
            uploads_root=Path(uploads_root) if uploads_root else None,
            skip_mmd=skip_mmd,
            skip_timeline=skip_timeline,
            dot_env=dot_env,
            **kwargs,
        )

    def _get_output_file(
        self,
        output_path: str | Path | None,
    ) -> Path:
        """Get the output file path."""
        if output_path:
            _output_path = Path(output_path)
            if _output_path.is_file():
                return _output_path
            if _output_path.is_dir():  # pragma: no branch
                filename = self._waldiez_file.stem
                return _output_path / f"{filename}.py"
        return self._waldiez_file.with_suffix(".py")

    @override
    def _run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        # pylint: disable=too-many-try-statements,broad-exception-caught
        mode = kwargs.get("mode", self.mode)
        if not isinstance(mode, str) or mode not in [
            "run",
            "debug",
        ]:  # pragma: no cover
            mode = "run"
        self.mode = mode  # type: ignore
        try:
            # Create sync subprocess runner
            runner = self._create_sync_subprocess_runner()

            # Run subprocess
            # noinspection PyTypeChecker
            runner.run_subprocess(self._waldiez_file, mode=self.mode)
            return self.read_from_output(output_file.parent)

        except BaseException as e:
            self.log.error("Error in sync subprocess execution: %s", e)
            return [
                {
                    "error": str(e),
                    "runner": "sync_subprocess",
                    "mode": self.mode,
                }
            ]

    @override
    async def a_run(
        self,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow asynchronously.

        Parameters
        ----------
        output_path : str | Path | None
            The output path, by default None.
        uploads_root : str | Path | None
            The runtime uploads root.
        structured_io : bool
            Whether to use structured IO instead of the default 'input/print'.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool
            Whether to skip generating the timeline JSON.
        dot_env : str | Path | None
            The path to the .env file, if any.
        **kwargs : Any
            Additional keyword arguments for the a_run method.

        Returns
        -------
        list[dict[str, Any]]
            The results of the run.
        """
        if dot_env is not None:  # pragma: no cover
            resolved = Path(dot_env).resolve()
            if resolved.is_file():
                WaldiezBaseRunner._dot_env_path = resolved
        temp_dir = Path.cwd()
        output_file = self._get_output_file(output_path)
        return await self._a_run(
            temp_dir=temp_dir,
            output_file=output_file,
            uploads_root=Path(uploads_root) if uploads_root else None,
            skip_mmd=skip_mmd,
            skip_timeline=skip_timeline,
            dot_env=dot_env,
            **kwargs,
        )

    @override
    async def _a_run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run workflow using async subprocess runner.

        Parameters
        ----------
        temp_dir : Path
            Temporary directory
        output_file : Path
            Output file path
        uploads_root : Path | None
            Uploads root directory
        skip_mmd : bool
            Skip mermaid diagram generation
        skip_timeline : bool
            Skip timeline generation
        **kwargs : Any
            Additional arguments

        Returns
        -------
        list[dict[str, Any]]
            Results of the workflow execution
        """
        mode = kwargs.get("mode", self.mode)
        if not isinstance(mode, str) or mode not in [
            "run",
            "debug",
        ]:  # pragma: no cover
            mode = "run"
        self.mode = mode  # type: ignore

        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            # Create async subprocess runner
            runner = self._create_async_subprocess_runner()

            # Run subprocess
            # noinspection PyTypeChecker
            await runner.run_subprocess(
                self._waldiez_file,
                mode=self.mode,
            )
            return await self.a_read_from_output(output_file.parent)

        except BaseException as e:
            self.log.error("Error in async subprocess execution: %s", e)
            return [
                {
                    "error": str(e),
                    "runner": "async_subprocess",
                    "mode": self.mode,
                }
            ]

    def provide_user_input(self, user_input: str) -> None:
        """Provide user input to the active subprocess runner.

        Parameters
        ----------
        user_input : str
            User input response
        """
        if self.async_runner and self.async_runner.is_running():
            asyncio.create_task(
                self.async_runner.provide_user_input(user_input)
            )
        if self.sync_runner and self.sync_runner.is_running():
            self.sync_runner.provide_user_input(user_input)

    async def a_provide_user_input(self, user_input: str) -> None:
        """Provide user input to the active subprocess runner (async version).

        Parameters
        ----------
        user_input : str
            User input response
        """
        if self.async_runner and self.async_runner.is_running():
            await self.async_runner.provide_user_input(user_input)
        if self.sync_runner and self.sync_runner.is_running():
            await asyncio.to_thread(
                self.sync_runner.provide_user_input, user_input
            )

    @override
    def stop(self) -> None:
        """Stop the workflow execution."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            super().stop()  # Set the base runner stop flag

            # Stop active subprocess runners
            if self.async_runner and self.async_runner.is_running():
                asyncio.create_task(self.async_runner.stop())
            if self.sync_runner and self.sync_runner.is_running():
                self.sync_runner.stop()
        except BaseException:
            pass
        try:
            SyncSubprocessRunner.gather()
        except BaseException:
            pass

    async def a_stop(self) -> None:
        """Stop the workflow execution (async version)."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            super().stop()  # Set the base runner stop flag

            # Stop active subprocess runners
            if self.async_runner and self.async_runner.is_running():
                await self.async_runner.stop()
            if self.sync_runner and self.sync_runner.is_running():
                await asyncio.to_thread(self.sync_runner.stop)
        except BaseException:
            pass
        try:
            await AsyncSubprocessRunner.gather()
        except BaseException:
            pass

    def is_subprocess_running(self) -> bool:
        """Check if a subprocess is currently running.

        Returns
        -------
        bool
            True if a subprocess is running, False otherwise
        """
        return (
            self.async_runner is not None and self.async_runner.is_running()
        ) or (self.sync_runner is not None and self.sync_runner.is_running())

    def get_subprocess_exit_code(self) -> int | None:
        """Get the exit code of the subprocess.

        Returns
        -------
        int | None
            Exit code if available, None otherwise
        """
        if self.sync_runner:
            return self.sync_runner.get_exit_code()
        return None

    def _cleanup_subprocess_runners(self) -> None:
        """Cleanup subprocess runner instances."""
        if self.async_runner and self.async_runner.is_running():
            asyncio.create_task(self.async_runner.stop())
        self.async_runner = None
        if self.sync_runner and self.sync_runner.is_running():
            self.sync_runner.stop()
        self.sync_runner = None

    @override
    def _after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        waldiez_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> None:
        """Actions to perform after running the flow.

        Parameters
        ----------
        results : list[dict[str, Any]]
            Results from the workflow execution.
        error : BaseException | None
            Optional error during the run.
        output_file : Path
            Output file path
        waldiez_file : Path
            The waldiez file used/dumped for the run.
        uploads_root : Path | None
            Uploads root directory
        temp_dir : Path
            Temporary directory
        skip_mmd : bool
            Skip mermaid diagram generation
        skip_timeline : bool
            Skip timeline generation
        """
        # Cleanup subprocess runners
        self._cleanup_subprocess_runners()

    @override
    async def _a_after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        waldiez_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> None:
        """Actions to perform after running the flow (async version).

        Parameters
        ----------
        results : list[dict[str, Any]]
            Results from the workflow execution.
        error : BaseException | None
            Optional error during the run.
        output_file : Path
            Output file path
        waldiez_file : Path
            The waldiez file used/dumped for the run.
        uploads_root : Path | None
            Uploads root directory
        temp_dir : Path
            Temporary directory
        skip_mmd : bool
            Skip mermaid diagram generation
        skip_timeline : bool
            Skip timeline generation
        """
        # Cleanup subprocess runners
        self._cleanup_subprocess_runners()

    @classmethod
    def create_with_callbacks(
        cls,
        waldiez_file: str | Path,
        on_output: Callable[[dict[str, Any]], None] | None = None,
        on_input_request: Callable[[str], None] | None = None,
        on_async_output: Callable[[dict[str, Any]], Any] | None = None,
        on_async_input_request: Callable[[str], Any] | None = None,
        **kwargs: Any,
    ) -> "WaldiezSubprocessRunner":
        """Create subprocess runner from waldiez file with callbacks.

        Parameters
        ----------
        waldiez_file : str | Path
            Path to waldiez file
        on_output : Callable[[dict[str, Any]], None] | None
            Sync output callback
        on_input_request : Callable[[str], None] | None
            Sync input request callback
        on_async_output : Callable[[dict[str, Any]], Any] | None
            Async output callback
        on_async_input_request : Callable[[str], Any] | None
            Async input request callback
        **kwargs : Any
            Additional arguments

        Returns
        -------
        WaldiezSubprocessRunner
            Configured subprocess runner
        """
        waldiez = Waldiez.load(waldiez_file)
        kwargs.pop("waldiez_file", None)
        return cls(
            waldiez=waldiez,
            on_output=on_output,
            on_input_request=on_input_request,
            on_async_output=on_async_output,
            on_async_input_request=on_async_input_request,
            waldiez_file=waldiez_file,
            **kwargs,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Exporting group chat utils.

Using the group patterns (no group manager agent)
"""

from waldiez.models import WaldiezGroupManager

from .common import get_event_handler_string


def export_group_chats(
    agent_names: dict[str, str],
    manager: WaldiezGroupManager,
    message: tuple[str, str] | None,
    tabs: int,
    is_async: bool,
) -> str:
    """Get the group chat string.

    Parameters
    ----------
    agent_names : dict[str, str]
        The agent names.
    manager : WaldiezGroupManager
        The group manager agent.
    message : str | None
        The initial message/messages arg and value for the chat
    tabs : int
        The number of tabs for indentation.
    is_async : bool
        Whether the chat is asynchronous.

    Returns
    -------
    tuple[str, str]
        The group chat string and the import string.
    """
    space = "    " * tabs
    run_group_chat: str = "run_group_chat"
    if is_async:
        run_group_chat = "await a_run_group_chat"
    manager_name = agent_names[manager.id]
    pattern_name = f"{manager_name}_pattern"
    content: str = f"{space}results = {run_group_chat}(" + "\n"
    content += f"{space}    pattern={pattern_name}," + "\n"
    if message:
        content += f"{space}    {message[0]}={message[1]},\n"
    content += f"{space}    max_rounds={manager.data.max_round},\n"
    if is_async:
        content += f"{space}    a_pause_event=a_pause_event,\n"
    else:
        content += f"{space}    pause_event=pause_event,\n"
    content += f"{space})\n"
    content += get_event_handler_string(space=space, is_async=is_async)
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=line-too-long

"""Extra requirements for agents."""

from .agent import WaldiezAgent
from .rag_user_proxy import WaldiezRagUserProxy


def get_retrievechat_extra_requirements(
    agents: list[WaldiezAgent],
) -> set[str]:
    """Get the retrievechat extra requirements.

    Parameters
    ----------
    agents : list[WaldiezAgent]
        The flow agents.

    Returns
    -------
    set[str]
        The retrievechat extra requirements.
    """
    # https://github.com/ag2ai/ag2/blob/main/pyproject.toml
    # with chromadb and sentence_transdormers relaxed
    rag_requirements: set[str] = {
        "protobuf==5.29.3",
        "chromadb>=0.5.23",
        "sentence_transformers",
        "pypdf",
        "ipython",
        "beautifulsoup4",
        "markdownify",
    }
    for agent in agents:
        if agent.is_rag_user and isinstance(agent, WaldiezRagUserProxy):
            # if not chroma, get the relevant db requirements
            db_type = agent.data.retrieve_config.vector_db
            if db_type == "pgvector":
                rag_requirements.update(
                    [
                        "pgvector>=0.2.5",
                        "psycopg[binary]>=3.2.4",
                    ]
                )
            elif db_type == "mongodb":
                rag_requirements.add("pymongo>=4.11")
            elif db_type == "qdrant":
                rag_requirements.update(["qdrant_client[fastembed]"])
    return rag_requirements


def get_captain_agent_extra_requirements() -> list[str]:
    """Get the captain agent extra requirements.

    Returns
    -------
    list[str]
        The captain agent extra requirements.
    """
    # https://github.com/ag2ai/ag2/blob/main/autogen/agentchat/contrib/captainagent/tools/requirements.txt  # noqa: E501
    tool_requirements = [
        "markdownify",
        "arxiv",
        "pymupdf",
        "wikipedia-api",
        "easyocr",
        "python-pptx",
        "openai-whisper",
        "scipy",
        # "pandas", also in agent_requirements below
        # "sentence-transformers", also in agent_requirements below
    ]
    agent_requirements = [
        "pandas",
        "chromadb",
        "sentence-transformers",
        "huggingface-hub",
    ]
    return tool_requirements + agent_requirements



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=too-many-public-methods

"""High-level storage manager for workspace and checkpoint operations."""

from __future__ import annotations

import os
import shutil
import sys
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from typing_extensions import Self

from .checkpoint import WaldiezCheckpoint, WaldiezCheckpointInfo
from .filesystem_storage import FilesystemStorage
from .protocol import Storage
from .utils import copy_results, get_root_dir, symlink


class StorageManager:
    """High-level storage manager to work with different storage backends."""

    _storage: Storage

    def __init__(
        self,
        storage: Storage | None = None,
        workspace_dir: Path | str | None = None,
    ) -> None:
        """
        Initialize the storage manager.

        Parameters
        ----------
        storage : Storage
            The Storage backend to use (defaults to FilesystemStorage)
        workspace_dir : Path | str | None
            Workspace directory (only used if storage is None)
        """
        if storage is None:
            if workspace_dir is None:
                workspace_dir = get_root_dir()
            self._storage = FilesystemStorage(workspace_dir)
        else:
            self._storage = storage

    @staticmethod
    def parse_checkpoint_arg(checkpoint_arg: str) -> tuple[str, int | None]:
        """Parse checkpoint arg.

        Parameters
        ----------
        checkpoint_arg : str
            The arg to parse.

        Returns
        -------
        tuple[str, int | None]
            The checkpoint id and an optional history index.
        """
        arg_parts = checkpoint_arg.split(":")
        history_index: int | None = None
        if len(arg_parts) == 2:
            try:
                history_index = int(arg_parts[1])
            except BaseException:  # pylint: disable=broad-exception-caught
                pass
        checkpoint_id = arg_parts[0]
        return checkpoint_id, history_index

    @staticmethod
    def default_root() -> Path:
        """Get the default global root.

        Returns
        -------
        Path
            The default global root.
        """
        return get_root_dir()

    @staticmethod
    def load_dict(json_file: Path) -> dict[str, Any]:
        """Load dict from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.

        Returns
        -------
        dict[str, Any]
            The loaded dict.
        """
        return FilesystemStorage.load_dict(json_file)

    @staticmethod
    def load_list(
        json_file: Path, fallback_dict_key: str
    ) -> list[dict[str, Any]]:
        """Load list from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.
        fallback_dict_key : str
            The key to get the list from if the loaded data is dict.

        Returns
        -------
        list[dict[str, Any]]
            The loaded list.
        """
        return FilesystemStorage.load_list(json_file, fallback_dict_key)

    @property
    def storage(self) -> Storage:
        """Get the underlying storage backend."""
        return self._storage

    @property
    def workspace_dir(self) -> Path:
        """Get the workspace directory."""
        return self._storage.workspace_dir

    # pylint: disable=too-many-locals,too-many-arguments
    def finalize(
        self,
        session_name: str,
        output_file: Path,
        tmp_dir: Path,
        *,
        metadata: dict[str, Any] | None = None,
        timestamp: datetime | None = None,
        link_root: Path | None = None,
        link_latest: bool = True,
        keep_tmp: bool = False,
        copy_into_subdir: str | None = None,
        promote_to_output: Iterable[str] = (
            "tree_of_thoughts.png",
            "reasoning_tree.json",
        ),
        ignore_names: Iterable[str] = (".cache", ".env"),
    ) -> tuple[Path, Path]:
        """Copy a run's temporary artifacts into a new checkpoint.

        Parameters
        ----------
        session_name : str
            Session name for the checkpoint.
        output_file : Path
            The path to the output.
        tmp_dir : Path
            Directory containing artifacts produced by the run.
        metadata : dict[str, Any] | None
            Optional metadata to store in the checkpoint.
        timestamp : datetime | None
            Optional checkpoint timestamp (defaults to now).
        link_root : Path | None
            Base directory for outward-facing links.
            Defaults to CWD/"waldiez_out".
        link_latest : bool
            Whether to update a `latest` link under link_root/session_name.
            Defaults to True.
        keep_tmp : bool
            If False, delete the tmp_dir after copying. Defaults to False.
        copy_into_subdir : str | None
            If set, copy artifacts into
            checkpoint/<copy_into_subdir> instead of the root.
        promote_to_output : Iterable[str]
            File names (exact matches) to also copy into output_dir.
        ignore_names : Iterable[str]
            Directory/file names to skip entirely.

        Returns
        -------
        tuple[Path, Path]
            (checkpoint_path, public_link_path)
            Path to the checkpoint in the workspace and the created public link.

        Raises
        ------
        FileNotFoundError
            If tmp_dir does not exist or is not a directory.
        """
        if not tmp_dir.exists() or not tmp_dir.is_dir():
            raise FileNotFoundError(
                f"tmp_dir does not exist or is not a directory: {tmp_dir}"
            )

        if timestamp is None:
            timestamp = datetime.now(timezone.utc)

        checkpoint_path = self._storage.save_checkpoint(
            session_name=session_name,
            state={},
            metadata=metadata or {},
            timestamp=timestamp,
        )

        target_dir = (
            (checkpoint_path / copy_into_subdir)
            if copy_into_subdir
            else checkpoint_path
        )
        target_dir.mkdir(parents=True, exist_ok=True)
        copy_results(
            temp_dir=tmp_dir,
            output_file=output_file,
            destination_dir=target_dir,
            promote_to_output=promote_to_output,
            ignore_names=ignore_names,
        )
        if link_root is None:
            link_root = Path.cwd() / "waldiez_out"

        session_out_dir = link_root / session_name
        session_out_dir.mkdir(parents=True, exist_ok=True)

        public_link_path = session_out_dir / checkpoint_path.name
        symlink(public_link_path, checkpoint_path, overwrite=True)

        if link_latest:
            latest_link = session_out_dir / "latest"
            tmp_latest = session_out_dir / ".latest.tmp"
            symlink(tmp_latest, checkpoint_path, overwrite=True)
            # pylint: disable=too-many-try-statements,broad-exception-caught
            # noinspection PyBroadException
            try:
                if sys.platform == "win32":
                    # Windows: remove target first
                    if os.path.exists(latest_link):
                        os.remove(latest_link)
                    os.rename(tmp_latest, latest_link)
                else:
                    # Unix-like: atomic replace
                    os.replace(tmp_latest, latest_link)
            except BaseException:
                pass
            finally:
                if tmp_latest.exists():
                    tmp_latest.unlink(missing_ok=True)

        if keep_tmp is False:
            shutil.rmtree(tmp_dir, ignore_errors=True)

        return checkpoint_path, public_link_path

    def save(
        self,
        session_name: str,
        state: dict[str, Any],
        metadata: dict[str, Any] | None = None,
    ) -> Path:
        """
        Save a checkpoint with automatic timestamp creation.

        Parameters
        ----------
        session_name : str
            The name of the session.
        state : dict[str, Any]
            The session state to store.
        metadata : dict[str, Any] | None
            Optional checkpoint metadata.

        Returns
        -------
        Path
            Path to the saved checkpoint
        """
        return self._storage.save_checkpoint(
            session_name=session_name, state=state, metadata=metadata
        )

    def get(
        self, session_name: str, timestamp: datetime | None = None
    ) -> WaldiezCheckpointInfo | None:
        """
        Get a checkpoint (latest by default).

        Parameters
        ----------
        session_name : str
            The name of the session
        timestamp : datetime | None
            Optional specific timestamp

        Returns
        -------
        WaldiezCheckpointInfo | None
            The loaded state data
        """
        return self._storage.get_checkpoint(
            session_name=session_name, timestamp=timestamp
        )

    def update(
        self,
        session_name: str,
        checkpoint: str | datetime,
        state: dict[str, Any],
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Update a checkpoint with new state and optionally new metadata.

        Parameters
        ----------
        session_name : str
            The name of the session
        checkpoint : str | datetime
            Specific timestamp for checkpoint.
        state : dict[str, Any]
            The new state to set.
        metadata : dict[str, Any]
            Optional new metadata to set.
        """
        if isinstance(checkpoint, str):
            # maybe a full
            dir_name = Path(checkpoint).name
            checkpoint_dt = WaldiezCheckpoint.parse_timestamp(dir_name)
            if not checkpoint_dt:
                return
        else:
            checkpoint_dt = checkpoint
        self._storage.save_checkpoint(
            session_name=session_name,
            state=state,
            metadata=metadata,
            timestamp=checkpoint_dt,
        )

    def load(
        self,
        info: WaldiezCheckpointInfo,
        history_index: int | None,
    ) -> WaldiezCheckpoint:
        """Load a checkpoint.

        Parameters
        ----------
        info: WaldiezCheckpointInfo
            The checkpoint info to load.
        history_index: int | None
            Optional history index to use.

        Returns
        -------
        dict[str, Any]
            The loaded checkpoint data.
        """
        return self._storage.load_checkpoint(info, history_index=history_index)

    def link(
        self,
        to: Path,
        session_name: str,
        timestamp: datetime | None = None,
    ) -> None:
        """
        Create a symlink to a checkpoint.

        Parameters
        ----------
        to: Path
            Where to create the symlink to.
        session_name : str
            The name of the session
        timestamp : datetime | None
            Optional specific timestamp
        """
        self._storage.link_checkpoint(
            to=to,
            session_name=session_name,
            timestamp=timestamp,
        )

    def checkpoints(
        self, session_name: str | None = None
    ) -> list[WaldiezCheckpointInfo]:
        """
        List available checkpoints.

        Parameters
        ----------
        session_name : str
            Optional filter by session

        Returns
        -------
        list[WaldiezCheckpointInfo]
            List of checkpoint information
        """
        return self._storage.list_checkpoints(session_name=session_name)

    def history(
        self, session_name: str, checkpoint_name: str | None = None
    ) -> dict[str, list[dict[str, Any]]]:
        """Get a session's checkpoints' history.

        Parameters
        ----------
        session_name : str
            The session to use.
        checkpoint_name : str | None
            Optional checkpoint/folder name to use

        Returns
        -------
        list[list[dict[str, Any]]]
            Each checkpoint's history.
        """
        entries: dict[str, list[dict[str, Any]]] = {}
        if checkpoint_name:
            checkpoint_dt = WaldiezCheckpoint.parse_timestamp(checkpoint_name)
            checkpoint_info = self._storage.get_checkpoint(
                session_name=session_name, timestamp=checkpoint_dt
            )
            if not checkpoint_info:
                return {"history": []}
            return {"history": checkpoint_info.checkpoint.history()}
        for checkpoint_info in self._storage.list_checkpoints(
            session_name=session_name
        ):
            checkpoint_history = checkpoint_info.checkpoint.history()
            if checkpoint_history:
                entries[
                    WaldiezCheckpoint.format_timestamp(
                        checkpoint_info.timestamp
                    )
                ] = checkpoint_history
        return entries

    def sessions(self) -> list[str]:
        """List available sessions.

        Returns
        -------
        list[str]
            The workspace sessions.
        """
        return self._storage.list_sessions()

    def delete_session(self, session_name: str) -> None:
        """Delete a session and all its checkpoints.

        Parameters
        ----------
        session_name : str
            The session to delete.
        """
        self.storage.delete_session(session_name)

    def delete(self, session_name: str, timestamp: datetime) -> None:
        """
        Delete a specific checkpoint.

        Args:
            session_name: Name of the session
            timestamp: Timestamp of the checkpoint
        """
        self._storage.delete_checkpoint(
            session_name=session_name, timestamp=timestamp
        )

    def cleanup(self, session_name: str, keep_count: int = 5) -> int:
        """
        Clean up old checkpoints.

        Parameters
        ----------
        session_name : str
            Name of the session
        keep_count : int
            Number of recent checkpoints to keep

        Returns
        -------
        int
            Number of checkpoints deleted
        """
        return self._storage.cleanup_old_checkpoints(
            session_name=session_name, keep_count=keep_count
        )

    def clean_broken_symlinks(self, session_name: str | None = None) -> int:
        """
        Clean up broken symlinks.

        Parameters
        ----------
        session_name : str | None
            If provided, clean only in that session's directory.
            If None, clean in all session directories.

        Returns
        -------
        int
            Number of broken symlinks removed.
        """
        # Only FilesystemStorage has this method currently
        if hasattr(self._storage, "clean_broken_symlinks"):
            return self._storage.clean_broken_symlinks(session_name)
        return 0

    @contextmanager
    def transaction(self) -> Generator[Self, None, None]:
        """
        Batch multiple operations together.

        Yields
        ------
        StorageManager
            Self

        Example
        -------
        with storage.transaction():
            storage.save("session1", state1)
            storage.save("session2", state2)
            storage.delete("session3", timestamp)
        """
        if hasattr(self._storage, "transaction"):
            with self._storage.transaction():
                yield self
        else:
            # No transaction support, just yield
            yield self

    def get_latest_checkpoint(
        self, session_name: str
    ) -> WaldiezCheckpointInfo | None:
        """Get information about the latest checkpoint for a session.

        Parameters
        ----------
        session_name : str
            Name of the session

        Returns
        -------
        WaldiezCheckpointInfo | None
            WaldiezCheckpoint info if found, None otherwise
        """
        checkpoints = self.checkpoints(session_name=session_name)
        if checkpoints:
            return checkpoints[0]  # Already sorted by timestamp desc
        return None

    def session_exists(self, session_name: str) -> bool:
        """
        Check if a session has any checkpoints.

        Parameters
        ----------
        session_name : str
            Name of the session

        Returns
        -------
        bool
            True if session has checkpoints.
        """
        checkpoints = self.checkpoints(session_name=session_name)
        return len(checkpoints) > 0



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=unused-argument
# pyright: reportImportCycles=false, reportUnusedParameter=false

"""Command handler for step-by-step execution."""

from typing import TYPE_CHECKING, Callable

from .step_by_step_models import (
    HELP_MESSAGE,
    WaldiezDebugError,
    WaldiezDebugStepAction,
)

if TYPE_CHECKING:
    # noinspection PyUnusedImports
    from .step_by_step_runner import WaldiezStepByStepRunner


# pylint: disable=too-few-public-methods
# noinspection PyUnusedLocal
class CommandHandler:
    """Handler for debug commands to reduce complexity in main runner."""

    def __init__(self, runner: "WaldiezStepByStepRunner"):
        self.runner = runner
        self._command_map: dict[
            str, Callable[[str | None], WaldiezDebugStepAction]
        ] = {
            "c": self._handle_continue,
            "s": self._handle_step,
            "r": self._handle_run,
            "q": self._handle_quit,
            "i": self._handle_info,
            "h": self._handle_help,
            "st": self._handle_stats,
            "ab": self._handle_add_breakpoint,
            "rb": self._handle_remove_breakpoint,
            "lb": self._handle_list_breakpoints,
            "cb": self._handle_clear_breakpoints,
        }

    def handle_command(self, command_line: str) -> WaldiezDebugStepAction:
        """Handle a command line input.

        Parameters
        ----------
        command_line : str
            The command line input to handle.

        Returns
        -------
        WaldiezDebugStepAction
            The action to take for the command.
        """
        if not command_line or not command_line.strip():
            return self._handle_step(None)  # Enter = step
        parts = command_line.strip().split(maxsplit=1)
        if not parts or not parts[0]:
            return self._handle_unknown(None)
        command = parts[0].lower()
        args = parts[1] if len(parts) > 1 else None

        handler = self._command_map.get(command, self._handle_unknown)
        return handler(args)

    def _handle_continue(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.step_mode = True
        return WaldiezDebugStepAction.CONTINUE

    def _handle_step(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.step_mode = True
        return WaldiezDebugStepAction.STEP

    def _handle_run(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.step_mode = False
        return WaldiezDebugStepAction.RUN

    def _handle_quit(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.set_stop_requested()
        return WaldiezDebugStepAction.QUIT

    def _handle_info(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.show_event_info()
        return WaldiezDebugStepAction.INFO

    def _handle_help(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.emit(HELP_MESSAGE)
        return WaldiezDebugStepAction.HELP

    def _handle_stats(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.show_stats()
        return WaldiezDebugStepAction.STATS

    def _handle_add_breakpoint(
        self, args: str | None
    ) -> WaldiezDebugStepAction:
        if args:
            self.runner.add_breakpoint(args)
            return WaldiezDebugStepAction.ADD_BREAKPOINT
        current_event = self.runner.current_event
        if current_event and hasattr(current_event, "type"):
            self.runner.add_breakpoint(getattr(current_event, "type", ""))
            return WaldiezDebugStepAction.ADD_BREAKPOINT
        # else:
        self.runner.emit(
            WaldiezDebugError(
                error=(
                    "No breakpoint specification provided "
                    "and no current event available"
                )
            )
        )
        return WaldiezDebugStepAction.ADD_BREAKPOINT

    def _handle_remove_breakpoint(
        self, args: str | None
    ) -> WaldiezDebugStepAction:
        if args:
            self.runner.remove_breakpoint(args)
            return WaldiezDebugStepAction.REMOVE_BREAKPOINT
        current_event = self.runner.current_event
        if current_event and hasattr(current_event, "type"):
            self.runner.remove_breakpoint(getattr(current_event, "type", ""))
            return WaldiezDebugStepAction.REMOVE_BREAKPOINT
        # else:
        self.runner.emit(
            WaldiezDebugError(
                error=(
                    "No breakpoint specification provided and "
                    "no current event available"
                )
            )
        )
        return WaldiezDebugStepAction.REMOVE_BREAKPOINT

    def _handle_list_breakpoints(
        self, args: str | None
    ) -> WaldiezDebugStepAction:
        self.runner.list_breakpoints()
        return WaldiezDebugStepAction.LIST_BREAKPOINTS

    def _handle_clear_breakpoints(
        self, args: str | None
    ) -> WaldiezDebugStepAction:
        self.runner.clear_breakpoints()
        return WaldiezDebugStepAction.CLEAR_BREAKPOINTS

    def _handle_unknown(self, args: str | None) -> WaldiezDebugStepAction:
        self.runner.emit(
            WaldiezDebugError(error="Unknown command. Use 'h' for help")
        )
        return WaldiezDebugStepAction.UNKNOWN



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=too-many-locals,unused-import,invalid-name
# pylint: disable=missing-function-docstring, missing-param-doc
# pyright: reportUnusedImport=false,reportConstantRedefinition=false
# pyright: reportUnusedParameter=false, reportCallInDefaultInitializer=false

"""CLI interface for Waldiez WebSocket server."""

import asyncio
import logging
import os
import re
import sys
import traceback
from pathlib import Path
from typing import Annotated, Any

import typer

HAS_WATCHDOG = False
try:
    from .reloader import FileWatcher  # noqa: F401

    HAS_WATCHDOG = True
except ImportError:
    pass

HAS_WEBSOCKETS = False
try:
    from .server import run_server

    HAS_WEBSOCKETS = True
except ImportError:
    # pylint: disable=missing-param-doc,missing-raises-doc
    # noinspection PyUnusedLocal
    async def run_server(*args: Any, **kwargs: Any) -> None:  # type: ignore
        """No WebSocket server available."""
        raise NotImplementedError("WebSocket server is not available.")


DEFAULT_WORKSPACE_DIR = Path.cwd()
DEFAULT_WS_PORT = 8765


def setup_logging(verbose: bool = False) -> None:
    """Set up logging configuration.

    Parameters
    ----------
    verbose : bool
        Enable verbose logging
    """
    level = logging.DEBUG if verbose else logging.INFO
    format_str = (
        "%(asctime)s - %(name)s - %(filename)s:%(lineno)d"
        " -%(levelname)s - %(message)s"
    )

    logging.basicConfig(
        level=level,
        format=format_str,
        handlers=[logging.StreamHandler(sys.stdout)],
    )

    # Set websockets logger to WARNING to reduce noise
    logging.getLogger("websockets").setLevel(logging.WARNING)


app = typer.Typer(
    name="waldiez-ws",
    help="Waldiez WebSocket server",
    add_completion=False,
    context_settings={
        "help_option_names": ["-h", "--help"],
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)


# noinspection PyBroadException
def _get_ws_port() -> int:
    """Get the default ws server port to use."""
    vite_ws_config = os.environ.get("VITE_DEV_WS_URL", "")
    if not vite_ws_config:
        return DEFAULT_WS_PORT
    # VITE_DEV_WS_URL=ws://localhost:8765
    server_parts = vite_ws_config.rsplit(":", maxsplit=1)
    if not server_parts:
        return DEFAULT_WS_PORT
    try:
        return int(server_parts[-1])
    except Exception:  # pylint: disable=broad-exception-caught
        return DEFAULT_WS_PORT


# noinspection PyBroadException
@app.command()
def serve(
    host: Annotated[
        str, typer.Option(help="Server host address")
    ] = "localhost",
    port: Annotated[int, typer.Option(help="Server port")] = _get_ws_port(),
    max_clients: Annotated[
        int,
        typer.Option(
            "--max-clients", help="Maximum number of concurrent clients"
        ),
    ] = 1,
    allowed_origins: Annotated[
        list[str] | None,
        typer.Option(
            "--allowed-origin",
            help=(
                "Allowed origins for CORS (can be used multiple times). "
                "Supports regex patterns. "
                "Examples: 'https://example.com', '.*\\.mydomain\\.com'"
            ),
        ),
    ] = None,
    auto_reload: Annotated[
        bool,
        typer.Option(
            "--auto-reload",
            help=(
                "Enable auto-reload on file changes "
                "(requires: pip install watchdog)"
            ),
        ),
    ] = False,
    watch_dir: Annotated[
        list[Path] | None,
        typer.Option(
            "--watch-dir",
            help=(
                "Additional directories to watch for auto-reload "
                "(can be used multiple times)"
            ),
        ),
    ] = None,
    workspace_dir: Annotated[
        Path,
        typer.Option(
            "--workspace",
            help="Path to the workspace directory",
            resolve_path=True,
            dir_okay=True,
            file_okay=False,
        ),
    ] = DEFAULT_WORKSPACE_DIR,
    ping_interval: Annotated[
        float,
        typer.Option(
            "--ping-interval", help="WebSocket ping interval in seconds"
        ),
    ] = 20.0,
    ping_timeout: Annotated[
        float,
        typer.Option(
            "--ping-timeout", help="WebSocket ping timeout in seconds"
        ),
    ] = 20.0,
    max_size: Annotated[
        int, typer.Option("--max-size", help="Maximum message size in bytes")
    ] = 8388608,
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help="Enable verbose logging")
    ] = False,
) -> None:
    """Start Waldiez WebSocket server."""
    setup_logging(verbose)

    logger = logging.getLogger(__name__)

    # Convert watch directories to set
    watch_dirs: set[Path] | None = None
    if watch_dir:
        watch_dirs = set(watch_dir)

    compiled_origins: list[re.Pattern[str]] | None = None
    if allowed_origins:
        try:
            compiled_origins = [
                re.compile(pattern) for pattern in allowed_origins
            ]
        except re.error as e:
            typer.echo(f"Invalid regex pattern in allowed origins: {e}")
            sys.exit(1)

    # Server configuration
    server_config: dict[str, Any] = {
        "max_clients": max_clients,
        "allowed_origins": compiled_origins,
        "ping_interval": ping_interval,
        "ping_timeout": ping_timeout,
        "max_size": max_size,
    }
    if not HAS_WATCHDOG and auto_reload:
        msg = (
            "Auto-reload requires the 'watchdog' package. "
            "Please install it with: pip install watchdog"
        )
        typer.echo(msg)
        auto_reload = False
    logger.info("Starting Waldiez WebSocket server...")
    logger.info("Configuration:")
    logger.info("  Host: %s", host)
    logger.info("  Port: %d", port)
    logger.info("  Max clients: %d", max_clients)
    logger.info("  Allowed origins: %s", allowed_origins or ["*"])
    logger.info("  Auto-reload: %s", auto_reload)
    logger.info("  Workspace directory: %s", workspace_dir)

    if watch_dirs:
        logger.info("  Watch directories: %s", watch_dirs)

    try:
        asyncio.run(
            run_server(
                host=host,
                port=port,
                workspace_dir=workspace_dir,
                auto_reload=auto_reload,
                watch_dirs=watch_dirs,
                **server_config,
            )
        )
    except KeyboardInterrupt:
        logger.info("Server stopped by user")
    except Exception:  # pylint: disable=broad-exception-caught
        tb = traceback.format_exc()
        logger.error("Server error: %s", tb)
        sys.exit(1)


if __name__ == "__main__":
    app()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common utils for all models."""

from .ag2_version import get_autogen_version
from .base import WaldiezBase
from .date_utils import now
from .dict_utils import update_dict
from .handoff import (
    WaldiezAgentTarget,
    WaldiezContextBasedCondition,
    WaldiezContextBasedTransition,
    WaldiezContextStrLLMCondition,
    WaldiezDefaultCondition,
    WaldiezExpressionContextCondition,
    WaldiezGroupOrNestedTarget,
    WaldiezHandoff,
    WaldiezHandoffCondition,
    WaldiezHandoffTransition,
    WaldiezLLMBasedCondition,
    WaldiezLLMBasedTransition,
    WaldiezRandomAgentTarget,
    WaldiezSimpleTarget,
    WaldiezStringContextCondition,
    WaldiezStringLLMCondition,
    WaldiezTransitionAvailability,
    WaldiezTransitionTarget,
)
from .id_generator import get_id
from .method_utils import (
    check_function,
    gather_code_imports,
    generate_function,
    get_function,
    parse_code_string,
)
from .naming import (
    MAX_VARIABLE_LENGTH,
    get_valid_instance_name,
    get_valid_python_variable_name,
    safe_filename,
)
from .waldiez_version import get_waldiez_version

__all__ = [
    "MAX_VARIABLE_LENGTH",
    "WaldiezBase",
    "check_function",
    "update_dict",
    "gather_code_imports",
    "get_autogen_version",
    "get_function",
    "get_id",
    "get_valid_instance_name",
    "get_valid_python_variable_name",
    "get_waldiez_version",
    "generate_function",
    "now",
    "parse_code_string",
    "safe_filename",
    "WaldiezDefaultCondition",
    "WaldiezAgentTarget",
    "WaldiezContextBasedCondition",
    "WaldiezContextBasedTransition",
    "WaldiezContextStrLLMCondition",
    "WaldiezExpressionContextCondition",
    "WaldiezGroupOrNestedTarget",
    "WaldiezHandoff",
    "WaldiezHandoffCondition",
    "WaldiezHandoffTransition",
    "WaldiezLLMBasedCondition",
    "WaldiezLLMBasedTransition",
    "WaldiezRandomAgentTarget",
    "WaldiezSimpleTarget",
    "WaldiezStringContextCondition",
    "WaldiezStringLLMCondition",
    "WaldiezTransitionAvailability",
    "WaldiezTransitionTarget",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownMemberType=false, reportAttributeAccessIssue=false
# pyright: reportMissingTypeStubs=false, reportDeprecated=false
# pylint: disable=duplicate-code,too-few-public-methods
"""Run a waldiez flow.

The flow is first converted to an autogen flow with agents, chats, and tools.
We then chown to temporary directory, call the flow's `main()` and
return the results. Before running the flow, any additional environment
variables specified in the waldiez file are set.
"""

import asyncio
import traceback
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

from typing_extensions import override

from waldiez.models.waldiez import Waldiez

from .base_runner import WaldiezBaseRunner
from .events_mixin import EventsMixin
from .results_mixin import WaldiezRunResults

if TYPE_CHECKING:
    from autogen.agentchat import ConversableAgent  # type: ignore
    from autogen.events import BaseEvent  # type: ignore
    from autogen.messages import BaseMessage  # type: ignore


MESSAGES = {
    "workflow_starting": "<Waldiez> - Starting workflow...",
    "workflow_finished": "<Waldiez> - Workflow finished",
    "workflow_stopped": "<Waldiez> - Workflow stopped by user",
    "workflow_failed": "<Waldiez> - Workflow execution failed: {error}",
}


# noinspection StrFormat
class WaldiezStandardRunner(WaldiezBaseRunner):
    """Run a waldiez flow in a standard way."""

    def __init__(
        self,
        waldiez: Waldiez,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Waldiez manager."""
        super().__init__(
            waldiez,
            output_path=output_path,
            uploads_root=uploads_root,
            structured_io=structured_io,
            dot_env=dot_env,
            **kwargs,
        )
        self._event_count = 0
        self._processed_events = 0

    # pylint: disable=unused-argument
    @override
    def _run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez workflow."""
        # pylint: disable=import-outside-toplevel
        from autogen.io import IOStream  # type: ignore

        from waldiez.io import StructuredIOStream

        results_container: WaldiezRunResults = {
            "results": [],
            "exception": None,
            "completed": False,
        }
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            loaded_module = self._load_module(output_file, temp_dir)
            WaldiezBaseRunner._store_run_paths(
                tmp_dir=temp_dir, output_file=output_file
            )
            if self._stop_requested.is_set():
                self.log.debug(
                    "Execution stopped before AG2 workflow start (sync)"
                )
                return []
            # noinspection DuplicatedCode
            if self.structured_io:
                stream = StructuredIOStream(
                    uploads_root=uploads_root, is_async=False
                )
            else:
                stream = IOStream.get_default()
            EventsMixin.set_print_function(stream.print)
            EventsMixin.set_input_function(stream.input)
            EventsMixin.set_send_function(stream.send)
            self._output_dir = temp_dir
            self.print(MESSAGES["workflow_starting"])
            self.print(self.waldiez.info.model_dump_json())
            results = loaded_module.main(
                on_event=self._on_event,
                state_json=self.state_json,
            )
            results_container["results"] = results
            self.print(MESSAGES["workflow_finished"])
        except SystemExit:  # pragma: no cover
            self.log.debug("Execution stopped by user (sync)")
            self.print(MESSAGES["workflow_stopped"])
        except BaseException as e:  # pragma: no cover
            results_container["exception"] = e
            traceback.print_exc()
            self.print(MESSAGES["workflow_failed"].format(error=e))

        finally:
            results_container["completed"] = True
            WaldiezBaseRunner._remove_run_paths()
        return results_container["results"]

    def _on_event(
        self,
        event: Union["BaseEvent", "BaseMessage"],
        agents: list["ConversableAgent"],
    ) -> bool:
        """Process an event from the workflow."""
        self._event_count += 1
        if self._stop_requested.is_set():
            self.log.debug(
                "Execution stopped before AG2 workflow event processing (sync)"
            )
            return False
        try:
            EventsMixin.process_event(
                event, agents, output_dir=self._output_dir
            )
            self._processed_events += 1
        except SystemExit:  # pragma: no cover
            self.log.debug("Execution stopped by user (sync)")
            return False
        except Exception as e:
            if self._stop_requested.is_set():
                self.log.debug("Exception during stop, returning False")
                return False
            raise RuntimeError(
                f"Error processing event {event}: {e}\n{traceback.format_exc()}"
            ) from e
        return not self._stop_requested.is_set()

    async def _a_on_event(
        self,
        event: Union["BaseEvent", "BaseMessage"],
        agents: list["ConversableAgent"],
    ) -> bool:
        """Process an event from the workflow asynchronously."""
        self._event_count += 1
        if self._stop_requested.is_set():  # pragma: no cover
            self.log.debug(
                "Execution stopped before AG2 workflow event processing (async)"
            )
            return False
        try:
            await EventsMixin.a_process_event(
                event, agents, output_dir=self._output_dir
            )
            self._processed_events += 1
        except SystemExit:  # pragma: no cover
            self.log.debug("Execution stopped by user (async)")
            return False
        except Exception as e:
            if self._stop_requested.is_set():
                self.log.debug("Exception during stop, returning False")
                return False
            raise RuntimeError(
                f"Error processing event {event}: {e}\n{traceback.format_exc()}"
            ) from e
        return not self._stop_requested.is_set()

    # pylint: disable=too-complex
    @override
    async def _a_run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez workflow asynchronously."""

        # fmt: off
        async def _execute_workflow() -> list[dict[str, Any]]:
            # fmt: on
            """Execute the workflow in an async context."""
            # pylint: disable=import-outside-toplevel
            from autogen.io import IOStream

            from waldiez.io import StructuredIOStream

            results: list[dict[str, Any]]
            # pylint: disable=too-many-try-statements,broad-exception-caught
            try:
                loaded_module = self._load_module(output_file, temp_dir)
                await WaldiezBaseRunner._a_store_run_paths(
                    tmp_dir=temp_dir, output_file=output_file,
                )
                if self._stop_requested.is_set():  # pragma: no cover
                    msg = (
                        "Execution stopped before AG2 "
                        "workflow event processing (async)"
                    )
                    self.log.debug(msg)
                    return []
                # noinspection DuplicatedCode
                if self.structured_io:
                    stream = StructuredIOStream(
                        uploads_root=uploads_root, is_async=True
                    )
                else:
                    stream = IOStream.get_default()
                EventsMixin.set_print_function(stream.print)
                EventsMixin.set_input_function(stream.input)
                EventsMixin.set_send_function(stream.send)
                self._output_dir = temp_dir
                self.print(MESSAGES["workflow_starting"])
                self.print(self.waldiez.info.model_dump_json())
                results = await loaded_module.main(
                    on_event=self._a_on_event,
                    state_json=self.state_json,
                )
                self.print(MESSAGES["workflow_finished"])
            except SystemExit:  # pragma: no cover
                self.log.debug("Execution stopped by user (async)")
                self.print(MESSAGES["workflow_stopped"])
                return []
            except Exception as e:  # pragma: no cover
                self.print(MESSAGES["workflow_failed"].format(error=str(e)))
                raise RuntimeError(
                    f"Error loading workflow: {e}\n{traceback.format_exc()}"
                ) from e
            finally:
                WaldiezBaseRunner._remove_run_paths()
            return results

        # Create cancellable task
        task = asyncio.create_task(_execute_workflow())

        # Monitor for stop requests
        # pylint: disable=too-many-try-statements
        try:
            while not task.done():
                if self._stop_requested.is_set():
                    task.cancel()
                    self.log.debug("Execution stopped by user (async)")
                    break
                await asyncio.sleep(0.1)
            # Return the task result when completed
            return await task

        except asyncio.CancelledError:
            self.log.debug("Execution cancelled (async)")
            return []



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Linting comments to include in the generated code."""

import math


def split_linter_comment(
    prefix: str,
    rules: list[str],
    max_lines: int = 3,
) -> str:
    """Split linter comment.

    Parameters
    ----------
    prefix : str
        The prefix for the comment, e.g., "# pylint: disable=" or "# pyright: ".
    rules : list[str]
        The list of linter rules to include in the comment.
    max_lines : int, optional
        The maximum number of lines to split the comment into, by default 3.

    Returns
    -------
    str
        The formatted comment string with the rules split into lines.
    """
    # Calculate minimum number of rules per line to not exceed max_lines
    rules_per_line = max(1, math.ceil(len(rules) / max_lines))
    # pylint: disable=inconsistent-quotes
    lines = [
        f"{prefix}{','.join(rules[i : i + rules_per_line])}"
        for i in range(0, len(rules), rules_per_line)
    ]
    return "\n".join(lines) + "\n"


PYLINT_RULES = [
    "broad-exception-caught",
    "f-string-without-interpolation",
    "invalid-name",
    "import-error",
    "import-outside-toplevel",
    "inconsistent-quotes",
    "line-too-long",
    "missing-function-docstring",
    "missing-param-doc",
    "missing-return-doc",
    "no-member",
    "pointless-string-statement",
    "too-complex",
    "too-many-arguments",
    "too-many-locals",
    "too-many-try-statements",
    "ungrouped-imports",
    "unnecessary-lambda-assignment",
    "unknown-option-value",
    "unused-argument",
    "unused-import",
    "unused-variable",
]
# including basedpyright specific (not pyright only) rules
PYRIGHT_RULES = [
    "reportArgumentType",
    "reportAttributeAccessIssue",
    "reportCallInDefaultInitializer",
    "reportConstantRedefinition",
    "reportDeprecated",
    "reportDuplicateImport",
    "reportMissingTypeStubs",
    "reportOperatorIssue",
    "reportOptionalMemberAccess",
    "reportPossiblyUnboundVariable",
    "reportUnreachable",
    "reportUnusedImport",
    "reportUnknownArgumentType",
    "reportUnknownMemberType",
    "reportUnknownLambdaType",
    "reportUnnecessaryIsInstance",
    "reportUnusedParameter",
    "reportUnusedVariable",
    "reportUnknownVariableType",
]

MYPY_RULES = [
    "import-untyped",
    "no-redef",
    "unused-ignore",
    "import-not-found",
]

FLAKE8_RULES = ["E501"]


def get_flake8_ignore_comment(rules: list[str] | None = None) -> str:
    """Get the flake8 ignore comment string.

    Parameters
    ----------
    rules : Optional[list[str]], optional
        The flake8 rules to ignore, by default None.

    Returns
    -------
    str
        The flake8 ignore comment string.

    Example
    -------
    ```python
    >>> get_flake8_ignore_comment(["E501", "F401"])

    # flake8: noqa: E501, F401
    ```
    """
    if not rules:
        rules = FLAKE8_RULES
    prefix = "# flake8: noqa: "
    output = ", ".join(rules)
    return prefix + output + "\n"


def get_pylint_ignore_comment(rules: list[str] | None = None) -> str:
    """Get the pylint ignore comment string.

    Parameters
    ----------
    rules : Optional[list[str]], optional
        The pylint rules to ignore, by default None.

    Returns
    -------
    str
        The pylint ignore comment string.

    Example
    -------
    ```python
    >>> get_pylint_ignore_comment(True, ["invalid-name", "line-too-long"])

    # pylint: disable=invalid-name, line-too-long
    ```
    """
    if not rules:
        rules = PYLINT_RULES
    prefix = "# pylint: disable="
    return split_linter_comment(
        prefix,
        rules,
        max_lines=3,
    )


def get_pyright_ignore_comment(rules: list[str] | None = None) -> str:
    """Get the pyright ignore comment string.

    Parameters
    ----------
    rules : Optional[list[str]], optional
        The pyright rules to ignore, by default None.

    Returns
    -------
    str
        The pyright ignore comment string.

    Example
    -------
    ```python
    >>> get_pyright_ignore_comment(
    ...     True,
    ...     ["reportUnusedImport", "reportMissingTypeStubs"]
    ... )

    # pyright: reportUnusedImport=false, reportMissingTypeStubs=false
    ```
    """
    if not rules:
        rules = PYRIGHT_RULES
    prefix = "# pyright: "
    output = split_linter_comment(
        prefix,
        [f"{rule}=false" for rule in rules],
        max_lines=3,
    )
    return output


def get_mypy_ignore_comment(rules: list[str] | None = None) -> str:
    """Get the mypy ignore comment string.

    Parameters
    ----------
    rules : Optional[list[str]], optional
        The mypy rules to ignore, by default None.

    Returns
    -------
    str
        The mypy ignore comment string.

    Example
    -------
    ```python
    >>> get_mypy_ignore_comment(["import-untyped", "no-redef"])

    # mypy: disable-error-code="import-untyped,no-redef"
    ```
    """
    if rules is None:
        rules = MYPY_RULES
    if not rules:
        return "# type: ignore\n"
    prefix = "# mypy: disable-error-code="
    content = ", ".join(rules)
    return prefix + f'"{content}"\n'
